"""Utility functions for type-safe API model creation.

This module provides helper functions to create API model objects from dictionaries,
ensuring type compatibility with mypy.
"""

from __future__ import annotations

from typing import Optional
from typing import cast

from spectrafit.api.moessbauer_model import AmplitudeAPI
from spectrafit.api.moessbauer_model import BackgroundAPI
from spectrafit.api.moessbauer_model import FwhmlAPI
from spectrafit.api.moessbauer_model import HyperfineFieldAPI
from spectrafit.api.moessbauer_model import IsomerShiftAPI
from spectrafit.api.moessbauer_model import QuadrupoleSplittingAPI


def create_amplitude_api(
    data: dict[str, float | bool | None],
) -> AmplitudeAPI:
    """Create an AmplitudeAPI instance from a dictionary.

    Args:
        data: Dictionary with amplitude parameters

    Returns:
        AmplitudeAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return AmplitudeAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )


def create_isomershift_api(
    data: dict[str, float | bool | None],
) -> IsomerShiftAPI:
    """Create an IsomerShiftAPI instance from a dictionary.

    Args:
        data: Dictionary with isomer shift parameters

    Returns:
        IsomerShiftAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return IsomerShiftAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )


def create_fwhml_api(
    data: dict[str, float | bool | None],
) -> FwhmlAPI:
    """Create a FwhmlAPI instance from a dictionary.

    Args:
        data: Dictionary with FWHM parameters

    Returns:
        FwhmlAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return FwhmlAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )


def create_background_api(
    data: dict[str, float | bool | None],
) -> BackgroundAPI:
    """Create a BackgroundAPI instance from a dictionary.

    Args:
        data: Dictionary with background parameters

    Returns:
        BackgroundAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return BackgroundAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )


def create_quadrupolesplitting_api(
    data: dict[str, float | bool | None],
) -> QuadrupoleSplittingAPI:
    """Create a QuadrupoleSplittingAPI instance from a dictionary.

    Args:
        data: Dictionary with quadrupole splitting parameters

    Returns:
        QuadrupoleSplittingAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return QuadrupoleSplittingAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )


def create_hyperfinefield_api(
    data: dict[str, float | bool | None],
) -> HyperfineFieldAPI:
    """Create a HyperfineFieldAPI instance from a dictionary.

    Args:
        data: Dictionary with hyperfine field parameters

    Returns:
        HyperfineFieldAPI: Properly instantiated model
    """
    vary = True
    expr = None

    if "vary" in data:
        vary_value = data.get("vary")
        vary = bool(vary_value) if vary_value is not None else True

    if "expr" in data:
        expr_value = data.get("expr")
        expr = cast("Optional[str]", expr_value)

    return HyperfineFieldAPI(
        max=data.get("max"),
        min=data.get("min"),
        vary=vary,
        value=data.get("value"),
        expr=expr,
    )



"""Reference model for the API of the Mössbauer models distributions."""

from __future__ import annotations

from pydantic import BaseModel
from pydantic import Field


__description__ = "Lmfit expression for explicit dependencies."


# Define base parameter classes locally to avoid circular imports
class AmplitudeAPI(BaseModel):
    """Definition of the amplitude of the models distributions."""

    max: float | None = Field(default=None, description="Maximum amplitude.")
    min: float | None = Field(default=None, description="Minimum amplitude.")
    vary: bool = Field(default=True, description="Vary the amplitude.")
    value: float | None = Field(default=None, description="Initial Amplitude value.")
    expr: str | None = Field(default=None, description=__description__)


class CenterAPI(BaseModel):
    """Definition of the center of the models distributions."""

    max: float | None = Field(default=None, description="Maximum center.")
    min: float | None = Field(default=None, description="Minimum center.")
    vary: bool = Field(default=True, description="Vary the center.")
    value: float | None = Field(default=None, description="Initial Center value.")
    expr: str | None = Field(default=None, description=__description__)


class IsomerShiftAPI(BaseModel):
    """Definition of the isomer shift (δ) parameter for Mössbauer spectra.

    The isomer shift is a measure of the electron density at the nucleus.
    Typically reported in mm/s.
    """

    max: float | None = Field(
        default=None,
        description="Maximum isomer shift value (mm/s).",
    )
    min: float | None = Field(
        default=None,
        description="Minimum isomer shift value (mm/s).",
    )
    vary: bool = Field(
        default=True,
        description="Vary the isomer shift parameter during fitting.",
    )
    value: float | None = Field(
        default=None,
        description="Initial isomer shift value (mm/s).",
    )
    expr: str | None = Field(default=None, description=__description__)


class FwhmlAPI(BaseModel):
    """Definition of the FWHM Lorentzian of the models distributions."""

    max: float | None = Field(
        default=None,
        description="Maximum Full Width Half Maximum of the Lorentzian Distribution.",
    )
    min: float | None = Field(
        default=None,
        description="Minimum Full Width Half Maximum of the Lorentzian Distribution.",
    )
    vary: bool = Field(
        default=True,
        description="Vary the Full Width Half Maximum of the Lorentzian Distribution.",
    )
    value: float | None = Field(
        default=None,
        description="Initial Full Width Half Maximum of "
        "the Lorentzian Distribution value.",
    )
    expr: str | None = Field(default=None, description=__description__)


class QuadrupoleSplittingAPI(BaseModel):
    """Definition of the Quadrupole Splitting parameter for Mössbauer distributions."""

    max: float | None = Field(
        default=None,
        description="Maximum Quadrupole Splitting value.",
    )
    min: float | None = Field(
        default=None,
        description="Minimum Quadrupole Splitting value.",
    )
    vary: bool = Field(
        default=True,
        description="Vary the Quadrupole Splitting parameter during fitting.",
    )
    value: float | None = Field(
        default=None,
        description="Initial Quadrupole Splitting value.",
    )
    expr: str | None = Field(default=None, description=__description__)


class HyperfineFieldAPI(BaseModel):
    """Definition of the Hyperfine Field parameter for Mössbauer distributions."""

    max: float | None = Field(
        default=None,
        description="Maximum Hyperfine Field value.",
    )
    min: float | None = Field(
        default=None,
        description="Minimum Hyperfine Field value.",
    )
    vary: bool = Field(
        default=True,
        description="Vary the Hyperfine Field parameter during fitting.",
    )
    value: float | None = Field(
        default=None,
        description="Initial Hyperfine Field value.",
    )
    expr: str | None = Field(default=None, description=__description__)


class BackgroundAPI(BaseModel):
    """Definition of the Background parameter for Mössbauer distributions."""

    max: float | None = Field(
        default=None,
        description="Maximum Background value.",
    )
    min: float | None = Field(
        default=None,
        description="Minimum Background value.",
    )
    vary: bool = Field(
        default=True,
        description="Vary the Background parameter during fitting.",
    )
    value: float | None = Field(
        default=None,
        description="Initial Background value.",
    )
    expr: str | None = Field(default=None, description=__description__)


class MoessbauerSingletAPI(BaseModel):
    r"""Definition of the Mössbauer Singlet distribution.

    A singlet is observed when the nucleus experiences no electric field gradient
    and no magnetic hyperfine field. It is characterized by a single absorption line
    at the isomer shift position.

    The mathematical expression for a Mössbauer singlet is:

    $$I(v) = I_0 - A \\cdot L(v-\\delta; \\Gamma)$$

    where:
    - $I(v)$ is the measured intensity at velocity $v$
    - $I_0$ is the background level
    - $A$ is the amplitude
    - $\\delta$ is the isomer shift
    - $\\Gamma$ is the linewidth (FWHM)
    - $L$ is the Lorentzian function
    """

    amplitude: AmplitudeAPI = AmplitudeAPI()
    isomershift: IsomerShiftAPI = IsomerShiftAPI()
    fwhml: FwhmlAPI = FwhmlAPI()
    background: BackgroundAPI = BackgroundAPI()


class MoessbauerDoubletAPI(BaseModel):
    r"""Definition of the Mössbauer Doublet distribution.

    A doublet is observed when the nucleus experiences an electric field gradient
    but no magnetic field. It is characterized by two absorption lines
    symmetrically positioned around the isomer shift position.

    The mathematical expression for a Mössbauer doublet is:

    $$
    I(v) = I_0 - \frac{A}{2} \cdot L(v-\delta-\frac{\Delta E_Q}{2}; \Gamma)
      - \frac{A}{2} \cdot L(v-\delta+\frac{\Delta E_Q}{2}; \Gamma)
    $$

    where:
    - $I(v)$ is the measured intensity at velocity $v$
    - $I_0$ is the background level
    - $A$ is the amplitude
    - $\delta$ is the isomer shift
    - $\Delta E_Q$ is the quadrupole splitting
    - $\Gamma$ is the linewidth (FWHM)
    - $L$ is the Lorentzian function
    """

    amplitude: AmplitudeAPI = AmplitudeAPI()
    isomershift: IsomerShiftAPI = IsomerShiftAPI()
    fwhml: FwhmlAPI = FwhmlAPI()
    quadrupolesplitting: QuadrupoleSplittingAPI = QuadrupoleSplittingAPI()
    background: BackgroundAPI = BackgroundAPI()


class MoessbauerSextetAPI(BaseModel):
    r"""Definition of the Mössbauer Sextet distribution.

    A sextet is observed when the nucleus experiences a magnetic field.
    It is characterized by six absorption lines due to magnetic splitting
    of the nuclear energy levels.

    The mathematical expression for a simplified Mössbauer sextet is:

    $$
    I(v) = I_0 - \sum_{i=1}^{6} \frac{A}{6} \cdot
    L(v-\delta-p_i \cdot B_{hf} + \frac{\Delta E_Q}{2}; \Gamma)
    $$

    where:
    - $I(v)$ is the measured intensity at velocity $v$
    - $I_0$ is the background level
    - $A$ is the amplitude
    - $\delta$ is the isomer shift
    - $B_{hf}$ is the hyperfine field
    - $\Delta E_Q$ is the quadrupole splitting
    - $p_i$ are the relative positions of the six lines
    - $\Gamma$ is the linewidth (FWHM)
    - $L$ is the Lorentzian function
    """

    amplitude: AmplitudeAPI = AmplitudeAPI()
    isomershift: IsomerShiftAPI = IsomerShiftAPI()
    fwhml: FwhmlAPI = FwhmlAPI()
    hyperfinefield: HyperfineFieldAPI = HyperfineFieldAPI()
    quadrupolesplitting: QuadrupoleSplittingAPI = QuadrupoleSplittingAPI()
    background: BackgroundAPI = BackgroundAPI()


class MoessbauerOctetAPI(BaseModel):
    r"""Definition of the Mössbauer Octet distribution.

    An octet is observed for magnetic materials with higher nuclear spin states,
    resulting in eight absorption lines due to complex hyperfine interactions.

    The mathematical expression for a simplified Mössbauer octet is:

    $$
    I(v) = I_0 - \sum_{i=1}^{8} \frac{A}{8} \cdot L(v-\delta-p_i \cdot B_{hf} +
    \frac{\Delta E_Q}{2}; \Gamma)
    $$

    where:
    - $I(v)$ is the measured intensity at velocity $v$
    - $I_0$ is the background level
    - $A$ is the amplitude
    - $\delta$ is the isomer shift
    - $B_{hf}$ is the hyperfine field
    - $\Delta E_Q$ is the quadrupole splitting
    - $p_i$ are the relative positions of the eight lines
    - $\Gamma$ is the linewidth (FWHM)
    - $L$ is the Lorentzian function
    """

    amplitude: AmplitudeAPI = AmplitudeAPI()
    isomershift: IsomerShiftAPI = IsomerShiftAPI()
    fwhml: FwhmlAPI = FwhmlAPI()
    hyperfinefield: HyperfineFieldAPI = HyperfineFieldAPI()
    quadrupolesplitting: QuadrupoleSplittingAPI = QuadrupoleSplittingAPI()
    background: BackgroundAPI = BackgroundAPI()


class MagneticAngles(BaseModel):
    """Parameters for magnetic field orientation."""

    theta: float = Field(default=0.0, description="Angle in radians")
    phi: float = Field(default=0.0, description="Angle in radians")


class EfgParameters(BaseModel):
    """Parameters for electric field gradient (EFG)."""

    vzz: float = Field(
        default=1e21,
        description="Principal component of EFG tensor (V/m²)",
    )
    eta: float = Field(default=0.0, description="Asymmetry parameter (dimensionless)")
    vary_vzz: bool = Field(default=False, description="Whether to vary vzz in fitting")
    vary_eta: bool = Field(default=False, description="Whether to vary eta in fitting")


class TemperatureParameters(BaseModel):
    """Parameters for temperature-dependent effects."""

    value: float = Field(default=300.0, description="Temperature in K")
    sod_shift: float = Field(
        default=0.0,
        description="Second-order Doppler shift (mm/s)",
    )
    vary_temp: bool = Field(
        default=False,
        description="Whether to vary temperature in fitting",
    )
    vary_sod: bool = Field(
        default=False,
        description="Whether to vary SOD shift in fitting",
    )


class DynamicParameter(BaseModel):
    """Parameter with bounds and variability for fitting."""

    value: float
    vary: bool = Field(
        default=False,
        description="Whether to vary this parameter in fitting",
    )
    min: float | None = Field(default=None, description="Minimum allowed value")
    max: float | None = Field(default=None, description="Maximum allowed value")



"""Convert the input and output files to the preferred file format."""

from __future__ import annotations

import argparse
import json

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar

import tomli_w
import yaml

from spectrafit.plugins.converter import Converter
from spectrafit.tools import read_input_file


if TYPE_CHECKING:
    from collections.abc import MutableMapping


class FileConverter(Converter):
    """Convert the input and output file to the preferred file format.

    !!! info "Supported file formats"

        Currently supported file formats:

        -[x] JSON
        -[x] YAML (YML)
        -[x] TOML (LOCK for the lock file)

    Attributes:
        choices (ClassVar[set[str]]): The choices for the file format.

    """

    choices: ClassVar[set[str]] = {"json", "yaml", "yml", "toml", "lock"}

    def get_args(self) -> dict[str, Any]:
        """Get the arguments from the command line.

        Returns:
            Dict[str, Any]: Return the input file arguments as a dictionary without
                additional information beyond the command line arguments.

        """
        parser = argparse.ArgumentParser(
            description="Converter for 'SpectraFit' input and output files.",
            usage="%(prog)s [options] infile",
        )
        parser.add_argument(
            "infile",
            type=Path,
            help="Filename of the 'SpectraFit' input or output file.",
        )
        parser.add_argument(
            "-f",
            "--file-format",
            help="File format for the conversion.",
            type=str,
            choices=self.choices,
        )
        parser.add_argument(
            "-e",
            "--export-format",
            help="File format for the export.",
            type=str,
            default="json",
            choices=self.choices,
        )
        return vars(parser.parse_args())

    @staticmethod
    def convert(infile: Path, file_format: str) -> MutableMapping[str, Any]:
        """Convert the input file to the output file.

        Args:
            infile (Path): The input file as a path object.
            file_format (str): The output file format.

        Raises:
            ValueError: If the input file format is not supported.

        Returns:
            MutableMapping[str, Any] : The converted file as a dictionary.

        """
        if file_format not in FileConverter.choices:
            msg = f"The input file format '{file_format}' is not supported."
            raise ValueError(msg)

        return read_input_file(infile)

    def save(self, data: Any, fname: Path, export_format: str) -> None:
        """Save the converted file.

        Raises:
            ValueError: If the input file format is identical with the output format.
            ValueError: If the output file format is not supported.

        Args:
            data (Any): The converted file as a dictionary.
            fname (Path): The input file as a path object.
            export_format (str): The output file format.

        """
        if fname.suffix[1:] == export_format:
            msg = (
                f"The input file suffix '{fname.suffix[1:]}' is similar to the"
                f" output file format '{export_format}'."
                "Please use a different output file suffix."
            )
            raise ValueError(
                msg,
            )

        if export_format not in self.choices:
            msg = f"The output file format '{export_format}' is not supported."
            raise ValueError(
                msg,
            )

        if export_format == "json":
            with fname.with_suffix(f".{export_format}").open(
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(data, f, indent=4)
        elif export_format in {"yaml", "yml"}:
            with fname.with_suffix(f".{export_format}").open(
                "w",
                encoding="utf-8",
            ) as f:
                yaml.dump(data, f, default_flow_style=False)
        elif export_format in {"toml", "lock"}:
            with fname.with_suffix(f".{export_format}").open("wb+") as f:
                tomli_w.dump(dict(**data), f)

    def __call__(self) -> None:
        """Run the converter via cmd commands."""
        args = self.get_args()
        self.save(
            data=self.convert(infile=args["infile"], file_format=args["file_format"]),
            fname=args["infile"],
            export_format=args["export_format"],
        )


def command_line_runner() -> None:
    """Run the converter from the command line."""
    FileConverter()()



"""SpectraFit, fast command line tool for fitting data.

!!! info "About Versioning"

    SpectraFit uses [Semantic Versioning](https://semver.org/).

!!! warning "About Python Versions"

    Currently, SpectraFit only supports Python 3.9 and above. Soon, Python 3.9
    will be deprecated in favor of Python 3.10 and above, see also
    [Release Schedule](https://devguide.python.org/versions/#end-of-life-branches).
"""

from __future__ import annotations

import sys
import warnings

from typing import Literal


PYTHON_END_OF_LIFE: tuple[Literal[3], Literal[9]] = (3, 9)
PACKAGE_LIFECYCLE_NOTICE = (
    "SpectraFit v2.0 is currently in development. "
    "SpectraFit v1.x is in end-of-life maintenance mode and will receive "
    "only critical hotfixes."
)

warnings.warn(
    PACKAGE_LIFECYCLE_NOTICE,
    FutureWarning,
    stacklevel=2,
)

if sys.version_info[:2] == PYTHON_END_OF_LIFE:
    version_str = f"{PYTHON_END_OF_LIFE[0]}.{PYTHON_END_OF_LIFE[1]}"
    warnings.warn(
        f"Support for Python {version_str} is approaching its end-of-life."
        " Please consider upgrading to Python 3.10 or newer."
        " For more details, see:"
        "https://devguide.python.org/versions/#end-of-life-branches.",
        DeprecationWarning,
        stacklevel=2,
    )

__version__ = "1.5.0"



"""Reference model for the API of the Jupyter Notebook interface."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field


class XAxisAPI(BaseModel):
    """Defintion of the X-Axis of the plotly figure."""

    name: str = Field(
        default="Incident Energy",
        description="Name of the x-axis of the plot.",
    )
    unit: str | None = Field(
        default="eV",
        description="Name of the x-axis units of the plot.",
    )


class YAxisAPI(BaseModel):
    """Defintion of the Y-Axis of the plotly figure."""

    name: str = Field(
        default="Emission Energy",
        description="Name of the y-axis of the plot.",
    )
    unit: str | None = Field(
        default="eV",
        description="Name of the y-axis units of the plot.",
    )


class ZAxisAPI(BaseModel):
    """Defintion of the Z-Axis of the plotly figure."""

    name: str = Field(
        default="Intensity",
        description="Name of the z-axis of the plot.",
    )
    unit: str | None = Field(
        default="a.u.",
        description="Name of the z-axis units of the plot.",
    )


class MainTitleAPI(BaseModel):
    """Defintion of the main title of the plotly figure."""

    rixs: str = Field(default="RIXS", description="Name of the RIXS plot.")
    xes: str = Field(default="XES", description="Name of the XES plot.")
    xas: str = Field(default="XAS", description="Name of the XAS plot.")


class SizeRatioAPI(BaseModel):
    """Defintion of the size ratio of the plotly figure."""

    size: tuple[int, int] = Field(
        default=(500, 500),
        description="Basic size of the plots in pixels.",
    )
    ratio_rixs: tuple[float, float] = Field(
        default=(2, 2),
        description="Ratio of the RIXS plot.",
    )
    ratio_xes: tuple[float, float] = Field(
        default=(3, 1),
        description="Ratio of the XES plot.",
    )
    ratio_xas: tuple[float, float] = Field(
        default=(3, 1),
        description="Ratio of the XAS plot.",
    )


class RIXSModelAPI(BaseModel):
    """Definition of the RIXS model."""

    incident_energy: Any = Field(
        ...,
        description="Incident energy values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    emission_energy: Any = Field(
        ...,
        description="Emission energy values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    rixs_map: Any = Field(
        ...,
        description="RIXS map values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    model_config = ConfigDict(arbitrary_types_allowed=True)


class RIXSPlotAPI(BaseModel):
    """Definition of the plotly figure."""

    incident_energy: Any = Field(
        ...,
        description="Incident energy values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    emission_energy: Any = Field(
        ...,
        description="Emission energy values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    emission_intensity: Any = Field(
        ...,
        description="Emission intensity values (RIXS), which has to be a 2D array.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    energy_loss: Any | None = Field(
        default=None,
        description="Energy loss values.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    energy_loss_intensity: Any | None = Field(
        default=None,
        description="Energy loss intensity values (XES), which has to be a 2D array.",
    )  # Should be NDArray[np.float64] but that's not supported for 3.8
    x_axis: XAxisAPI = XAxisAPI()
    y_axis: YAxisAPI = YAxisAPI()
    z_axis: ZAxisAPI = ZAxisAPI()
    main_title: MainTitleAPI = MainTitleAPI()
    size_ratio: SizeRatioAPI = SizeRatioAPI()
    model_config = ConfigDict(arbitrary_types_allowed=True)



"""Definition of the data file model."""

from __future__ import annotations

from pathlib import Path
from typing import Callable

from pydantic import BaseModel
from pydantic import Field
from pydantic.functional_validators import field_validator


class DataFileAPI(BaseModel):
    """Definition of a data file."""

    skiprows: int | None = Field(
        default=None,
        description="Number of lines to skip at the beginning of the file.",
    )
    skipfooter: int = Field(
        ...,
        description="Number of lines to skip at the end of the file.",
    )
    delimiter: str = Field(
        ...,
        description="Delimiter to use.",
    )
    comment: str | None = Field(
        default=None,
        description="Comment marker to use.",
    )
    names: Callable[[Path, str], list[str] | None] | None = Field(
        default=None,
        description="Column names can be provided by list of strings or a function",
    )
    header: int | list[str] | None = Field(
        default=None,
        description="Column headers to use.",
    )
    file_suffixes: list[str] = Field(
        ...,
        description="File suffixes to use.",
    )

    @field_validator("delimiter")
    @classmethod
    def check_delimiter(cls, v: str) -> str | None:
        """Check if the delimiter is valid."""
        if v in {" ", "\t", ",", ";", "|", r"\s+"}:
            return v
        msg = f" {v} is not a valid delimiter."
        raise ValueError(msg)

    @field_validator("comment")
    @classmethod
    def check_comment(cls, v: str) -> str | None:
        """Check if the comment marker is valid."""
        if v is None or v in {"#", "%"}:
            return v
        msg = f" {v} is not a valid comment marker."
        raise ValueError(msg)



"""Transform the input data to a CSV file."""

from __future__ import annotations

import argparse
import re

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import pandas as pd

from spectrafit.api.file_model import DataFileAPI
from spectrafit.plugins.converter import Converter


if TYPE_CHECKING:
    from collections.abc import MutableMapping


def get_athena_column(fname: Path, comment: str = "#") -> list[str] | None:
    """Get the header of the file.

    Args:
        fname (Path): The file name of the data file.
        comment (str, optional): The comment marker. Defaults to "#".

    Returns:
        Optional[List[str]]: The column names of the data file as a list.

    """
    with fname.open(encoding="utf-8") as f:
        data = f.read()
        lines = data.splitlines()
        return next(
            (
                lines[i - 1].split(comment)[-1].split()
                for i, line in enumerate(lines)
                if re.match(r"^\s*\d", line)
            ),
            None,
        )


athena_format = DataFileAPI(
    skiprows=None,
    skipfooter=0,
    delimiter=r"\s+",
    comment="#",
    names=get_athena_column,
    file_suffixes=[".nor"],
)


txt_format = DataFileAPI(
    skiprows=None,
    skipfooter=0,
    delimiter=r"\s+",
    header=0,
    file_suffixes=[".txt", ".dat", ".out"],
)


@dataclass(frozen=True)
class DataFormats:
    """Data formats."""

    ATHENA = athena_format
    TXT = txt_format


# Remove the private attributes
choices = {
    choice for choice in list(DataFormats.__dict__.keys()) if not choice.startswith("_")
}

choices_export = {"csv", "txt", "dat", "out"}


class DataConverter(Converter):
    """Convert the data files to a CSV file.

    !!! info "Supported file formats"

        Currently supported file formats:

        -[x] ATHENA
        -[x] TXT
        -[ ] more to come

        `DataConverter` class can be also used in the Jupyter notebook.
    """

    def get_args(self) -> dict[str, Any]:
        """Get the arguments from the command line.

        Returns:
            Dict[str, Any]: Return the input file arguments as a dictionary without
                additional information beyond the command line arguments.

        """
        parser = argparse.ArgumentParser(
            description="Converter for 'SpectraFit' from data files to CSV files.",
            usage="%(prog)s [options] infile",
        )
        parser.add_argument(
            "infile",
            type=Path,
            help="Filename of the data file to convert.",
        )
        parser.add_argument(
            "-f",
            "--file-format",
            help="File format for the conversion.",
            type=str,
            choices=choices,
        )
        parser.add_argument(
            "-e",
            "--export-format",
            help="File format for the export.",
            type=str,
            default="csv",
            choices=choices_export,
        )
        return vars(parser.parse_args())

    @staticmethod
    def convert(infile: Path, file_format: str) -> MutableMapping[str, Any]:
        """Convert the input file to the target file format.

        Args:
            infile (Path): Input file as a path object.
            file_format (str): Target file format.

        Raises:
            ValueError: If the file format is not supported.

        Returns:
            MutableMapping[str, Any]: The converted data as a MutableMapping[str, Any],
                which belongs to DataFrame.

        """
        if file_format.upper() not in choices:
            msg = f"File format '{file_format}' is not supported."
            raise ValueError(msg)

        if callable(DataFormats.__dict__[file_format].names):
            names = DataFormats.__dict__[file_format].names(infile)
        else:
            names = DataFormats.__dict__[file_format].names
        DataFormats.__dict__[file_format].names = names

        return pd.read_csv(
            infile,
            **DataFormats.__dict__[file_format].dict(exclude={"file_suffixes"}),
        )

    def save(self, data: Any, fname: Path, export_format: str) -> None:
        """Save the converted data to a CSV file.

        Raises:
            ValueError: If the export format is not supported.

        Args:
            data (Any): The converted data, which is a pandas DataFrame.
            fname (Path): The file name of the data file.
            export_format (str): The file format of the exported file.

        """
        if export_format.lower() not in choices_export:
            msg = f"Export format '{export_format}' is not supported."
            raise ValueError(msg)
        data.to_csv(fname.with_suffix(f".{export_format}"), index=False)

    def __call__(self) -> None:
        """Run the converter."""
        args = self.get_args()
        self.save(
            data=self.convert(
                args["infile"],
                args["file_format"],
            ),
            fname=args["infile"],
            export_format=args["export_format"],
        )


def command_line_runner() -> None:
    """Run the converter from the command line."""
    DataConverter()()



"""Reference model for the API of the Jupyter Notebook report."""

from __future__ import annotations

from collections.abc import Hashable
from typing import Any

from dtale import __version__ as dtale_version
from emcee import __version__ as emcee_version
from itables import __version__ as itables_version
from lmfit import __version__ as lmfit_version
from numdifftools import __version__ as numdifftools_version
from numpy import __version__ as numpy_version
from pandas import __version__ as pandas_version
from plotly import __version__ as plotly_version
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import __version__ as pydantic_version
from scipy import __version__ as scipy_version
from sklearn import __version__ as sklearn_version
from statsmodels import __version__ as statsmodels_version

from spectrafit.api.cmd_model import DescriptionAPI
from spectrafit.api.tools_model import DataPreProcessingAPI


class CreditsAPI(BaseModel):
    """Credits API model."""

    dtale: str = f"dtale v{dtale_version}"
    emcee: str = f"emcee v{emcee_version}"
    itables: str = f"itables v{itables_version}"
    lmfit: str = f"lmfit v{lmfit_version}"
    numdifftools: str = f"numdifftools v{numdifftools_version}"
    numpy: str = f"numpy v{numpy_version}"
    pandas: str = f"pandas v{pandas_version}"
    plotly: str = f"plotly v{plotly_version}"
    pydantic: str = f"pydantic v{pydantic_version}"
    scipy: str = f"scipy v{scipy_version}"
    sklearn: str = f"sklearn v{sklearn_version}"
    statsmodels: str = f"statsmodels v{statsmodels_version}"


class FitMethodAPI(BaseModel):
    """Fit method API model."""

    global_fitting: bool | int = Field(
        default=False,
        description="Fitting in the global fashion",
    )
    confidence_interval: bool | dict[str, Any] = Field(
        ...,
        description="Settings for the confidence interval calculation",
    )
    configurations: dict[str, Any] = Field(
        ...,
        description="Settings for the fitting configuration",
    )
    settings_solver_models: dict[str, Any] = Field(
        ...,
        description="Settings for the solver models including minimizer and optimizer",
    )


class InputAPI(BaseModel):
    """Input API for the report endpoint."""

    description: DescriptionAPI = DescriptionAPI()
    credits: CreditsAPI = CreditsAPI()
    initial_model: list[dict[str, dict[str, dict[str, Any]]]] = Field(
        ...,
        description="Initial model for the fit",
    )
    method: FitMethodAPI = Field(
        ...,
        description="Fitting method with optional including of confidence interval",
    )
    pre_processing: DataPreProcessingAPI = Field(..., description="Data pre-processing")


class SolverAPI(BaseModel):
    """Solver API for the report endpoint."""

    goodness_of_fit: dict[str, float] = Field(..., description="Goodness of fit")
    regression_metrics: dict[str, list[Any]] = Field(
        ...,
        description="Regression metrics",
    )
    descriptive_statistic: dict[str, list[Any]] = Field(
        ...,
        description="Descriptive statistic",
    )
    linear_correlation: dict[str, list[Any]] = Field(
        ...,
        description="Linear correlation",
    )
    component_correlation: dict[str, dict[str, Any]] = Field(
        default={},
        description="Linear correlation of each attribute of components. if possible",
    )
    confidence_interval: dict[str, Any] = Field(
        default={},
        description="Confidence interval, if possible",
    )
    covariance_matrix: dict[str, dict[str, Any]] = Field(
        default={},
        description="Covariance matrix, if possible",
    )
    variables: dict[str, dict[str, Any]] = Field(
        ...,
        description="Variables with their initial, optimized and optional error values",
    )
    errorbars: dict[str, Any] = Field(
        default={},
        description="Error bar comment if values reach initial value or boundary",
    )
    computational: dict[str, Any] = Field(
        ...,
        description="Computational information like number of function evaluations",
    )


class OutputAPI(BaseModel):
    """Output API for the report endpoint."""

    df_org: dict[Hashable, Any] = Field(
        ...,
        description="DataFrame of the original data via 'records' orient",
    )
    df_fit: dict[Hashable, Any] = Field(
        ...,
        description="DataFrame of the fitted data via 'records' orient",
    )
    df_pre: dict[Hashable, Any] = Field(
        default={},
        description="DataFrame of the pre-processed data via 'records' orient",
    )
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ReportAPI(BaseModel):
    """Definition of the report model."""

    input: InputAPI = Field(
        ...,
        description="Input data for the report.",
    )
    solver: SolverAPI = Field(
        ...,
        description="Solver data for the report.",
    )
    output: OutputAPI = Field(
        ...,
        description="Output data for the report.",
    )



"""Abstract base class for the converter plugins."""

from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from pathlib import Path


class Converter(ABC):
    """Abstract base class for the converter plugin.

    The abstract base class is used to define the interface for the converter plugins:

    - get_args: Get the arguments from the command line.
    - convert: Convert the input file to the output file.
    - __call__: Call the converter plugin.

    Currently used for:

    - Convertion of the input file.
    - Convertion of the output file.
    """

    @abstractmethod
    def get_args(self) -> dict[str, Any]:
        """Get the arguments from the command line.

        Returns:
            Dict[str, Any]: Return the input file arguments as a dictionary without
                 additional information beyond the command line arguments.

        Raises:
            ValueError: If the output file format is not supported.

        """

    @staticmethod
    @abstractmethod
    def convert(infile: Path, file_format: str) -> MutableMapping[str, Any]:
        """Convert the input file to the target file format.

        It is an abstract method and must be implemented in the derived class.

        Args:
            infile (Path): Input file as a path object.
            file_format (str): Target file format.

        Returns:
            MutableMapping[str, Any]: Converted file as a dictionary.

        """

    @abstractmethod
    def save(self, data: Any, fname: Path, export_format: str) -> None:
        """Save the data to the target file format.

        Args:
            data (Any): Data to save.
            fname (Path): Filename of the target file.
            export_format (str): Target file format.

        """

    @abstractmethod
    def __call__(self) -> None:
        """Call the converter plugin."""



"""Jupyter Notebook plugin for SpectraFit."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
import pandas as pd
import plotly.express as px
import tomli_w

from IPython.display import display
from IPython.display import display_markdown
from dtale import show as dtale_show
from itables import show as itables_show
from plotly.subplots import make_subplots

from spectrafit.api.cmd_model import DescriptionAPI
from spectrafit.api.models_model import ConfIntervalAPI
from spectrafit.api.notebook_model import ColorAPI
from spectrafit.api.notebook_model import FnameAPI
from spectrafit.api.notebook_model import FontAPI
from spectrafit.api.notebook_model import GridAPI
from spectrafit.api.notebook_model import LegendAPI
from spectrafit.api.notebook_model import MetricAPI
from spectrafit.api.notebook_model import PlotAPI
from spectrafit.api.notebook_model import ResidualAPI
from spectrafit.api.notebook_model import RunAPI
from spectrafit.api.notebook_model import XAxisAPI
from spectrafit.api.notebook_model import YAxisAPI
from spectrafit.api.report_model import FitMethodAPI
from spectrafit.api.report_model import InputAPI
from spectrafit.api.report_model import OutputAPI
from spectrafit.api.report_model import ReportAPI
from spectrafit.api.report_model import SolverAPI
from spectrafit.api.tools_model import ColumnNamesAPI
from spectrafit.api.tools_model import DataPreProcessingAPI
from spectrafit.api.tools_model import SolverModelsAPI
from spectrafit.models.builtin import SolverModels
from spectrafit.tools import PostProcessing
from spectrafit.tools import PreProcessing
from spectrafit.tools import exclude_none_dictionary
from spectrafit.tools import transform_nested_types
from spectrafit.utilities.transformer import list2dict


if TYPE_CHECKING:
    from plotly.graph_objects import Figure

# Constants
MIN_DATAFRAME_COLUMNS = 2  # Minimum number of columns required in a dataframe


class DataFrameDisplay:
    """Class for displaying a dataframe in different ways."""

    def df_display(self, df: pd.DataFrame, mode: str | None = None) -> Any | None:
        """Call the DataframeDisplay class.

        !!! info "About `df_display`"

            This function is used to display a dataframe in two different ways.

            1. Regular display mode:
                1. Via `IPython.display` for regular sliced displaying of the dataframe
                   in the notebook.
                2. Via `IPython.display` as Markdown for regular displaying of the
                    complete dataframe in the notebook.
            2. Interactive display mode:
                1. Via `itables` for interactive displaying of the dataframe in the
                    notebook, which allows for sorting, filtering, and jumping. For
                    more information see [itables](https://github.com/mwouts/itables).
                2. Via `dtale` for interactive displaying of the dataframe in the
                    notebook, which allows advanced data analysis of the dataframe in
                    an external window. For more information see
                    [dtale](https://github.com/man-group/dtale).

        Args:
            df (pd.DataFrame): Dataframe to display.
            mode (str, Optional): Display mode. Defaults to None.

        Raises:
            ValueError: Raises ValueError if mode of displaying is not supported.

        Returns:
            Optional[Any]: Returns the dtale object for plotting in the Jupyter
                 notebook, if mode is `dtale`.

        """
        if mode == "regular":
            self.regular_display(df=df)
        elif mode == "markdown":
            self.markdown_display(df=df)
        elif mode == "interactive":
            self.interactive_display(df=df)
        elif mode == "dtale":
            return self.dtale_display(df=df)
        elif mode is not None:
            msg = (
                f"Invalid mode: {mode}. "
                "Valid modes are: regular, interactive, dtale, markdown."
            )
            raise ValueError(
                msg,
            )
        return None

    @staticmethod
    def regular_display(df: pd.DataFrame) -> None:
        """Display the dataframe in a regular way.

        Args:
            df (pd.DataFrame): Dataframe to display.

        """
        display(df)

    @staticmethod
    def interactive_display(df: pd.DataFrame) -> None:
        """Display the dataframe in an interactive way.

        Args:
            df (pd.DataFrame): Dataframe to display.

        """
        itables_show(df)

    @staticmethod
    def dtale_display(df: pd.DataFrame) -> Any:
        """Display the dataframe in a dtale way.

        Args:
            df (pd.DataFrame): Dataframe to display.

        Returns:
            Any: Returns the dtale object for plotting in the Jupyter notebook.

        """
        return dtale_show(df)

    @staticmethod
    def markdown_display(df: pd.DataFrame) -> None:
        """Display the dataframe in a markdown way.

        Args:
            df (pd.DataFrame): Dataframe to display.

        """
        display_markdown(df.to_markdown(), raw=True)


class DataFramePlot:
    """Class to plot a dataframe."""

    def plot_2dataframes(
        self,
        args_plot: PlotAPI,
        df_1: pd.DataFrame,
        df_2: pd.DataFrame | None = None,
    ) -> None:
        """Plot two dataframes.

        !!! info "About the plot"

            The plot is a combination of two plots. The first plot can be the
            residual plot of a fit or the _modified_ data. The second plot can be the
            fit or the original data.

        !!! missing "`line_dash_map`"

            Currently, the `line_dash_map` is not working, and the dash is not
            plotted. This is likely due to the columns not being labeled in the
            dataframe.

        Args:
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.
            df_1 (pd.DataFrame): First dataframe to plot, which will generate
                a fit plot with residual plot. The ratio is 70% to 20% with
                10% space in between.
            df_2 (Optional[pd.DataFrame], optional): Second optional dataframe to
                plot for comparison. In this case, the ratio between the first
                and second plot will be the same. Defaults to None.

        """
        if df_2 is None:
            fig = self._plot_single_dataframe(args_plot, df_1)
        else:
            fig = self._plot_two_dataframes(args_plot, df_1, df_2)

        fig.show(
            config={
                "toImageButtonOptions": {
                    "format": "png",
                    "filename": "plot_of_2_dataframes",
                    "scale": 4,
                },
            },
        )

    def _plot_single_dataframe(self, args_plot: PlotAPI, df: pd.DataFrame) -> Figure:
        """Plot a single dataframe with residuals."""
        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            shared_yaxes=True,
            vertical_spacing=0.05,
        )

        residual_fig = self._create_residual_plot(df, args_plot)
        fit_fig = self._create_fit_plot(df, args_plot)

        for trace in residual_fig["data"]:
            fig.add_trace(trace, row=1, col=1)
        for trace in fit_fig["data"]:
            fig.add_trace(trace, row=2, col=1)

        self._update_plot_layout(fig, args_plot, df_2_provided=False)
        return fig

    def _plot_two_dataframes(
        self,
        args_plot: PlotAPI,
        df_1: pd.DataFrame,
        df_2: pd.DataFrame,
    ) -> Figure:
        """Plot two dataframes for comparison."""
        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            shared_yaxes=True,
            vertical_spacing=0.05,
        )

        fig1 = px.line(df_1, x=args_plot.x, y=args_plot.y)
        fig2 = px.line(df_2, x=args_plot.x, y=args_plot.y)

        for trace in fig1["data"]:
            fig.add_trace(trace, row=1, col=1)
        for trace in fig2["data"]:
            fig.add_trace(trace, row=2, col=1)

        self._update_plot_layout(fig, args_plot, df_2_provided=True)
        return fig

    def _create_residual_plot(self, df: pd.DataFrame, args_plot: PlotAPI) -> Figure:
        """Create the residual plot."""
        return px.line(
            df,
            x=ColumnNamesAPI().energy,
            y=ColumnNamesAPI().residual,
            color_discrete_sequence=[args_plot.color.residual],
        )

    def _create_fit_plot(self, df: pd.DataFrame, args_plot: PlotAPI) -> Figure:
        """Create the fit plot."""
        y_columns = df.columns.drop(
            [ColumnNamesAPI().energy, ColumnNamesAPI().residual],
        )
        color_map = {
            ColumnNamesAPI().intensity: args_plot.color.intensity,
            ColumnNamesAPI().fit: args_plot.color.fit,
            **dict.fromkeys(
                y_columns.drop([ColumnNamesAPI().intensity, ColumnNamesAPI().fit]),
                args_plot.color.components,
            ),
        }
        line_dash_map = {
            ColumnNamesAPI().intensity: "solid",
            ColumnNamesAPI().fit: "longdash",
            **dict.fromkeys(
                y_columns.drop([ColumnNamesAPI().intensity, ColumnNamesAPI().fit]),
                "dash",
            ),
        }
        return px.line(
            df,
            x=ColumnNamesAPI().energy,
            y=y_columns,
            color_discrete_map=color_map,
            line_dash_map=line_dash_map,
        )

    def _update_plot_layout(
        self,
        fig: Figure,
        args_plot: PlotAPI,
        df_2_provided: bool,
    ) -> None:
        """Update the plot layout."""
        height = args_plot.size[1][0]
        self.update_layout_axes(fig, args_plot, height)

        xaxis_title = self.title_text(
            name=args_plot.xaxis_title.name,
            unit=args_plot.xaxis_title.unit,
        )
        yaxis_title = self.title_text(
            name=args_plot.yaxis_title.name,
            unit=args_plot.yaxis_title.unit,
        )

        fig.update_xaxes(title_text=xaxis_title, row=1, col=1)
        fig.update_xaxes(title_text=xaxis_title, row=2, col=1)

        if not df_2_provided:
            residual_title = self.title_text(
                name=args_plot.residual_title.name,
                unit=args_plot.residual_title.unit,
            )
            fig["layout"]["yaxis1"].update(domain=[0.8, 1])
            fig["layout"]["yaxis2"].update(domain=[0, 0.7])
            fig.update_yaxes(title_text=residual_title, row=1, col=1)
            fig.update_yaxes(title_text=yaxis_title, row=2, col=1)

            # Apply y-axis inversion if requested (for main spectrum)
            if args_plot.yaxis_title.invert:
                fig.update_yaxes(autorange="reversed", row=2, col=1)
        else:
            fig.update_yaxes(title_text=yaxis_title, row=1, col=1)
            fig.update_yaxes(title_text=yaxis_title, row=2, col=1)

            # Apply y-axis inversion if requested (for both plots in this layout)
            if args_plot.yaxis_title.invert:
                fig.update_yaxes(autorange="reversed", row=1, col=1)
                fig.update_yaxes(autorange="reversed", row=2, col=1)

    def plot_dataframe(self, args_plot: PlotAPI, df: pd.DataFrame) -> None:
        """Plot the dataframe according to the PlotAPI arguments.

        Args:
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.
            df (pd.DataFrame): Dataframe to plot.

        """
        fig = px.line(df, x=args_plot.x, y=args_plot.y)
        height = args_plot.size[1][0]
        self.update_layout_axes(fig, args_plot, height)

        fig.update_xaxes(
            title_text=self.title_text(
                name=args_plot.xaxis_title.name,
                unit=args_plot.xaxis_title.unit,
            ),
        )
        fig.update_yaxes(
            title_text=self.title_text(
                name=args_plot.yaxis_title.name,
                unit=args_plot.yaxis_title.unit,
            ),
        )
        fig.show(
            config={
                "toImageButtonOptions": {
                    "format": "png",
                    "filename": "plot_dataframe",
                    "scale": 4,
                },
            },
        )

    def plot_global_fit(self, args_plot: PlotAPI, df: pd.DataFrame) -> None:
        """Plot the global dataframe according to the PlotAPI arguments.

        Args:
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.
            df (pd.DataFrame): Dataframe to plot.

        """
        num_fits = df.columns.str.startswith(ColumnNamesAPI().fit).sum()
        for i in range(1, num_fits + 1):
            cols = [col for col in df.columns if col.endswith(f"_{i}")]
            cols.append(ColumnNamesAPI().energy)
            df_subset = df[cols].rename(
                columns={
                    f"{ColumnNamesAPI().intensity}_{i}": ColumnNamesAPI().intensity,
                    f"{ColumnNamesAPI().fit}_{i}": ColumnNamesAPI().fit,
                    f"{ColumnNamesAPI().residual}_{i}": ColumnNamesAPI().residual,
                },
            )
            self.plot_2dataframes(args_plot, df_subset)

    def plot_metric(
        self,
        args_plot: PlotAPI,
        df_metric: pd.DataFrame,
        bar_criteria: str | list[str],
        line_criteria: str | list[str],
    ) -> None:
        """Plot the metric according to the PlotAPI arguments.

        Args:
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.
            df_metric (pd.DataFrame): Metric dataframe to plot.
            bar_criteria (Union[str, List[str]]): Criteria to plot as bars.
            line_criteria (Union[str, List[str]]): Criteria to plot as lines.

        """
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig_bar = px.bar(
            df_metric,
            y=bar_criteria,
            color_discrete_sequence=args_plot.color.bars,
        )
        fig_line = px.line(
            df_metric,
            y=line_criteria,
            color_discrete_sequence=args_plot.color.lines,
        )
        fig_line.update_traces(mode="lines+markers", yaxis="y2")

        for trace in fig_bar.data:
            fig.add_trace(trace)
        for trace in fig_line.data:
            fig.add_trace(trace)

        fig.update_layout(xaxis_type="category")
        height = args_plot.size[1][1]
        self.update_layout_axes(fig, args_plot, height)

        fig.update_xaxes(
            title_text=self.title_text(
                name=args_plot.run_title.name,
                unit=args_plot.run_title.unit,
            ),
        )
        fig.update_yaxes(
            title_text=self.title_text(
                name=args_plot.metric_title.name_0,
                unit=args_plot.metric_title.unit_0,
            ),
            secondary_y=False,
        )
        fig.update_yaxes(
            title_text=self.title_text(
                name=args_plot.metric_title.name_1,
                unit=args_plot.metric_title.unit_1,
            ),
            secondary_y=True,
        )
        fig.show(
            config={
                "toImageButtonOptions": {
                    "format": "png",
                    "filename": "plot_metric",
                    "scale": 4,
                },
            },
        )

    def update_layout_axes(
        self,
        fig: Figure,
        args_plot: PlotAPI,
        height: int,
    ) -> Figure:
        """Update the layout of the plot.

        Args:
            fig (Figure): Figure to update.
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.
            height (int): Height of the plot.

        Returns:
            Figure: Updated figure.

        """
        fig.update_layout(
            title=args_plot.title,
            legend_title=args_plot.legend_title,
            legend=args_plot.legend.model_dump(),
            font=args_plot.font.model_dump(),
            showlegend=args_plot.show_legend,
            width=args_plot.size[0],
            height=height,
            paper_bgcolor=args_plot.color.paper,
            plot_bgcolor=args_plot.color.plot,
        )

        minor_ticks = self.get_minor(args_plot)

        fig.update_xaxes(
            minor=minor_ticks,
            gridcolor=args_plot.color.grid,
            linecolor=args_plot.color.line,
            zerolinecolor=args_plot.color.zero_line,
            color=args_plot.color.color,
        )
        fig.update_yaxes(
            minor=minor_ticks,
            gridcolor=args_plot.color.grid,
            linecolor=args_plot.color.line,
            zerolinecolor=args_plot.color.zero_line,
            color=args_plot.color.color,
        )
        return fig

    @staticmethod
    def title_text(name: str, unit: str | None = None) -> str:
        """Return the title text.

        Args:
            name (str): Name of the variable.
            unit (Optional[str], optional): Unit of the variable. Defaults to None.

        Returns:
            str: Title text.

        """
        return f"{name} [{unit}]" if unit else name

    def get_minor(self, args_plot: PlotAPI) -> dict[str, str | bool]:
        """Get the minor axis arguments.

        Args:
            args_plot (PlotAPI): PlotAPI object for the settings of the plot.

        Returns:
            Dict[str, Union[str, bool]]: Dictionary with the minor axis arguments.

        """
        return {
            "tickcolor": args_plot.color.ticks,
            "showgrid": args_plot.grid.show,
            "ticks": args_plot.grid.ticks,
            "griddash": args_plot.grid.dash,
        }


class ExportResults:
    """Class for exporting results as csv."""

    def export_df(self, df: pd.DataFrame, args: FnameAPI) -> None:
        """Export the dataframe as csv.

        Args:
            df (pd.DataFrame): Dataframe to export.
            args (FnameAPI): Arguments for the file export including the path, prefix,
                 and suffix.

        """
        df.to_csv(
            self.fname2path(
                fname=args.fname,
                prefix=args.prefix,
                suffix=args.suffix,
                folder=args.folder,
            ),
            index=False,
        )

    def export_report(self, report: dict[Any, Any], args: FnameAPI) -> None:
        """Export the results as toml file.

        Args:
            report (Dict[Any, Any]): Results as dictionary to export.
            args (FnameAPI): Arguments for the file export including the path, prefix,
                 and suffix.

        """
        with self.fname2path(
            fname=args.fname,
            prefix=args.prefix,
            suffix=args.suffix,
            folder=args.folder,
        ).open("wb+") as f:
            tomli_w.dump(report, f)

    @staticmethod
    def fname2path(
        fname: str,
        suffix: str,
        prefix: str | None = None,
        folder: str | None = None,
    ) -> Path:
        """Translate string to Path object.

        Args:
            fname (str): Filename
            suffix (str): Name of the suffix of the file.
            prefix (Optional[str], optional): Name of the prefix of the file. Defaults
                 to None.
            folder (Optional[str], optional): Folder, where it will be saved.
                 This folders will be created, if not exist. Defaults to None.

        Returns:
            Path: Path object of the file.

        """
        if prefix:
            fname = f"{prefix}_{fname}"
        _fname = Path(fname).with_suffix(f".{suffix}")
        if folder:
            Path(folder).mkdir(parents=True, exist_ok=True)
            _fname = Path(folder) / _fname
        return _fname


class SolverResults:
    """Class for storing the results of the solver."""

    def __init__(self, args_out: dict[str, Any]) -> None:
        """Initialize the SolverResults class.

        Args:
            args_out (Dict[str, Any]): Dictionary of SpectraFit settings and results.

        """
        self.args_out = args_out

    @property
    def settings_global_fitting(self) -> bool | int:
        """Global fitting settings.

        Returns:
            Union[bool, int]: Global fitting settings.

        """
        return self.args_out["global_"]

    @property
    def settings_configurations(self) -> dict[str, Any]:
        """Configure settings.

        Returns:
            Dict[str, Any]: Configuration settings.

        """
        return self.args_out["fit_insights"]["configurations"]

    @property
    def get_gof(self) -> dict[str, float]:
        """Get the goodness of fit values.

        Returns:
            Dict[str, float]: Goodness of fit values as dictionary.

        """
        return self.args_out["fit_insights"]["statistics"]

    @property
    def get_variables(self) -> dict[str, dict[str, float]]:
        """Get the variables of the fit.

        Returns:
            Dict[str, Dict[str, float]]: Variables of the fit.

        """
        return self.args_out["fit_insights"]["variables"]

    @property
    def get_errorbars(self) -> dict[str, float]:
        """Get the comments about the error bars of fit values.

        Returns:
            Dict[str, float]: Comments about the error bars as dictionary or dataframe.

        """
        return self.args_out["fit_insights"]["errorbars"]

    @property
    def get_component_correlation(self) -> dict[str, Any]:
        """Get the linear correlation of the components.

        Returns:
            Dict[str, Any]: Linear correlation of the components as dictionary.

        """
        return self.args_out["fit_insights"]["correlations"]

    @property
    def get_covariance_matrix(self) -> dict[str, Any]:
        """Get the covariance matrix.

        Returns:
            Dict[str, Any]: Covariance matrix as dictionary.

        """
        return self.args_out["fit_insights"]["covariance_matrix"]

    @property
    def get_regression_metrics(self) -> dict[str, Any]:
        """Get the regression metrics.

        Returns:
            Dict[str, Any]: Regression metrics as dictionary.

        """
        return self.args_out["regression_metrics"]

    @property
    def get_descriptive_statistic(self) -> dict[str, Any]:
        """Get the descriptive statistic.

        Returns:
            Dict[str, Any]: Descriptive statistic as dictionary of the spectra, fit, and
                 components as dictionary.

        """
        return self.args_out["descriptive_statistic"]

    @property
    def get_linear_correlation(self) -> dict[str, Any]:
        """Get the linear correlation.

        Returns:
            Dict[str, Any]: Linear correlation of the spectra, fit, and components
                 as dictionary.

        """
        return self.args_out["linear_correlation"]

    @property
    def get_computational(self) -> dict[str, Any]:
        """Get the computational time.

        Returns:
            Dict[str, Any]: Computational time as dictionary.

        """
        return self.args_out["fit_insights"]["computational"]

    @property
    def settings_conf_interval(self) -> bool | dict[str, Any]:
        """Confidence interval settings.

        Returns:
            Union[bool, Dict[str, Any]]: Confidence interval settings.

        """
        if isinstance(self.args_out["conf_interval"], dict):
            self.args_out["conf_interval"] = {
                key: value if value is not None else {}
                for key, value in self.args_out["conf_interval"].items()
            }
        return self.args_out["conf_interval"]

    @property
    def get_confidence_interval(self) -> dict[Any, Any]:
        """Get the confidence interval.

        Returns:
            Dict[Any, Any]: Confidence interval as dictionary with or without the
                    confidence interval results.

        """
        if self.args_out["conf_interval"] is False:
            return {}
        return self.args_out["confidence_interval"]

    @property
    def get_current_metric(self) -> pd.DataFrame:
        """Get the current metric.

        !!! note "About the regression metrics"

            For using the regression metrics, the `regression_metrics` must be averaged
            to merge the results of the different configurations together with the
            `goodness_of_fit` and `variables` results.

        Returns:
            pd.DataFrame: Current metric based on `regression_metrics` and
            `goodness_of_fit` as dataframe.

        """
        gof = {key: [value] for key, value in self.get_gof.items()}
        reg = {
            key: [np.average(val)]
            for key, val in zip(
                self.get_regression_metrics["index"],
                self.get_regression_metrics["data"],
            )
        }
        metric = {**gof, **reg}
        return pd.DataFrame(metric)


class ExportReport(SolverResults):
    """Class for exporting results as toml."""

    def __init__(
        self,
        description: DescriptionAPI,
        initial_model: list[dict[str, dict[str, dict[str, Any]]]],
        pre_processing: DataPreProcessingAPI,
        settings_solver_models: SolverModelsAPI,
        fname: FnameAPI,
        args_out: dict[str, Any],
        df_org: pd.DataFrame,
        df_fit: pd.DataFrame,
        df_pre: pd.DataFrame | None = None,
    ) -> None:
        """Initialize the ExportReport class.

        Args:
            description (DescriptionAPI): Description of the fit project.
            initial_model (List[Dict[str, Dict[str, Dict[str, Any]]]]): Initial model
                 for the fit.
            pre_processing (DataPreProcessingAPI): Data pre-processing settings.
            settings_solver_models (SolverModelsAPI): Solver models settings.
            fname (FnameAPI): Filename of the fit project including the path, prefix,
                 and suffix.
            args_out (Dict[str, Any]): Dictionary of SpectraFit settings and results.
            df_org (pd.DataFrame): Dataframe of the original data for performing
                 the fit.
            df_fit (pd.DataFrame): Dataframe of the final fit data.
            df_pre (Optional[pd.DataFrame], optional): Dataframe of the pre-processed data.
                 Defaults to None (empty DataFrame).

        """
        super().__init__(args_out=args_out)
        self.description = description
        self.initial_model = initial_model
        self.pre_processing = pre_processing
        self.settings_solver_models = settings_solver_models
        self.fname = fname

        self.df_org = df_org.to_dict(orient="list")
        self.df_fit = df_fit.to_dict(orient="list")
        if df_pre is None:  # pragma: no cover
            df_pre = pd.DataFrame()
        self.df_pre = df_pre.to_dict(orient="list")

    @property
    def make_input_contribution(self) -> InputAPI:
        """Make input contribution of the report.

        Returns:
            InputAPI: Input contribution of the report as class.

        """
        return InputAPI(
            description=self.description,
            initial_model=self.initial_model,
            pre_processing=self.pre_processing,
            method=FitMethodAPI(
                global_fitting=self.settings_global_fitting,
                confidence_interval=self.settings_conf_interval,
                configurations=self.settings_configurations,
                settings_solver_models=self.settings_solver_models.model_dump(
                    exclude_none=True,
                ),
            ),
        )

    @property
    def make_solver_contribution(self) -> SolverAPI:
        """Make solver contribution of the report.

        Returns:
            SolverAPI: Solver contribution of the report as class.

        """
        return SolverAPI(
            goodness_of_fit=self.get_gof,
            regression_metrics=self.get_regression_metrics,
            descriptive_statistic=self.get_descriptive_statistic,
            linear_correlation=self.get_linear_correlation,
            component_correlation=self.get_component_correlation,
            confidence_interval=self.get_confidence_interval,
            covariance_matrix=self.get_covariance_matrix,
            variables=self.get_variables,
            errorbars=self.get_errorbars,
            computational=self.get_computational,
        )

    @property
    def make_output_contribution(self) -> OutputAPI:
        """Make output contribution of the report.

        Returns:
            OutputAPI: Output contribution of the report as class.

        """
        return OutputAPI(df_org=self.df_org, df_fit=self.df_fit, df_pre=self.df_pre)

    def __call__(self) -> dict[str, Any]:
        """Get the complete report as dictionary.

        !!! info "About the report and `exclude_none_dictionary`"

            The report is generated by using the `ReportAPI` class, which is a
            `Pydantic`-definition of the report. The `Pydantic`-definition is
            converted to a dictionary by using the `.model_dump()` option of `Pydantic`.
            The `recursive_exclude_none` function is used to remove all `None` values
            from the dictionary, which are hidden in the nested dictionaries.

        Returns:
            Dict[str, Any]: Report as dictionary by using the `.dict()` option of
                 pydantic. `None` is excluded.

        """
        report = ReportAPI(
            input=self.make_input_contribution,
            solver=self.make_solver_contribution,
            output=self.make_output_contribution,
        ).model_dump(exclude_none=True)
        report = exclude_none_dictionary(report)
        return transform_nested_types(report)


class SpectraFitNotebook(DataFramePlot, DataFrameDisplay, ExportResults):
    """Jupyter Notebook plugin for SpectraFit."""

    args: dict[str, Any]
    global_: bool | int = False
    autopeak: bool = False
    df_fit: pd.DataFrame
    df_pre: pd.DataFrame = pd.DataFrame()
    df_metric: pd.DataFrame = pd.DataFrame()
    df_peaks: pd.DataFrame = pd.DataFrame()
    initial_model: list[dict[str, dict[str, dict[str, Any]]]]

    def __init__(  # noqa: C901
        self,
        df: pd.DataFrame,
        x_column: str,
        y_column: str | list[str],
        oversampling: bool = False,
        smooth: int = 0,
        shift: float = 0,
        energy_start: float | None = None,
        energy_stop: float | None = None,
        title: str | None = None,
        xaxis_title: XAxisAPI | None = None,
        yaxis_title: YAxisAPI | None = None,
        residual_title: ResidualAPI | None = None,
        metric_title: MetricAPI | None = None,
        run_title: RunAPI | None = None,
        legend_title: str = "Spectra",
        show_legend: bool = True,
        legend: LegendAPI | None = None,
        font: FontAPI | None = None,
        minor_ticks: bool = True,
        color: ColorAPI | None = None,
        grid: GridAPI | None = None,
        size: tuple[int, tuple[int, int]] = (800, (600, 300)),
        fname: str = "results",
        folder: str | None = None,
        description: DescriptionAPI | None = None,
    ) -> None:
        """Initialize the SpectraFitNotebook class.

        !!! info "About `Pydantic`-Definition"

            For being consistent with the `SpectraFit` class, the `SpectraFitNotebook`
            class refers to the `Pydantic`-Definition of the `SpectraFit` class.
            Currently, the following definitions are used:

            - `XAxisAPI`: Definition of the x-axis including units
            - `YAxisAPI`: Definition of the y-axis including units
            - `ResidualAPI`: Definition of the residual including units
            - `LegendAPI`: Definition of the legend according to `Plotly`
            - `FontAPI`: Definition of the font according to `Plotly`, which can be
                replaced by _built-in_ definitions
            - `ColorAPI`: Definition of the colors according to `Plotly`, which can be
                replace by _built-in_ definitions
            - `GridAPI`: Definition of the grid according to `Plotly`
            - `DescriptionAPI`: Definition of the description of the fit project

            All classes can be replaced by the corresponding `dict`-definition.

            ```python
            LegendAPI(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1)
            ```

            can be also

            ```python
            dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1)
            ```

        Args:
            df (pd.DataFrame): Dataframe with the data to fit.
            x_column (str): Name of the x column.
            y_column (Union[str, List[str]]): Name of the y column(s).
            oversampling (bool, optional): Activate the oversampling options.
                 Defaults to False.
            smooth (int, optional): Activate the smoothing functions setting an
                 `int>0`. Defaults to 0.
            shift (float, optional): Apply shift to the x-column. Defaults to 0.
            energy_start (Optional[float], optional): Energy start. Defaults to None.
            energy_stop (Optional[float], optional): Energy stop. Defaults to None.
            title (Optional[str], optional): Plot title. Defaults to None.
            xaxis_title (XAxisAPI, optional): X-Axis title. Defaults to XAxisAPI().
            yaxis_title (YAxisAPI, optional): Y-Axis title. Defaults to YAxisAPI().
            residual_title (ResidualAPI, optional): Residual title. Defaults to
                 ResidualAPI().
            metric_title (MetricAPI, optional): Metric title for both axes, bar and
                 line plot. Defaults to MetricAPI().
            run_title (RunAPI, optional): Run title. Defaults to RunAPI().
            legend_title (str, optional): Legend title. Defaults to "Spectra".
            show_legend (bool, optional): Show legend. Defaults to True.
            legend (LegendAPI, optional): Legend options. Defaults to LegendAPI().
            font (FontAPI, optional): Font options. Defaults to FontAPI().
            minor_ticks (bool, optional): Show minor ticks. Defaults to True.
            color (ColorAPI, optional): Color options. Defaults to ColorAPI().
            grid (GridAPI, optional): Grid options. Defaults to GridAPI().
            size (Tuple[int, Tuple[int, int]] , optional): Size of the fit- and metric-
                 plot. First width defines the fit, the second the metrics.
                 Defaults to (800, (600,300)).
            fname (str, optional): Filename of the export. Defaults to "results".
            folder (Optional[str], optional): Folder of the export. Defaults to None.
            description (DescriptionAPI, optional): Description of the data. Defaults
                 to DescriptionAPI()..


        Raises:
            ValueError: If the dataframe only contains one column.

        """
        self.x_column = x_column
        self.y_column = y_column

        if df.shape[1] < MIN_DATAFRAME_COLUMNS:
            msg = f"The dataframe must have {MIN_DATAFRAME_COLUMNS} or more columns."
            raise ValueError(msg)

        if isinstance(self.y_column, list):
            self.global_ = 1
            self.df = df[[self.x_column, *self.y_column]]
        else:
            self.df = df[[self.x_column, self.y_column]]
        self.df_org = self.df.copy()

        self.args_pre = DataPreProcessingAPI(
            oversampling=oversampling,
            energy_start=energy_start,
            energy_stop=energy_stop,
            smooth=smooth,
            shift=shift,
            column=list(self.df.columns),
        )
        # Initialize default value for description if None
        if description is None:
            description = DescriptionAPI()
        self.args_desc = description

        # Initialize default values for all API objects if None
        if xaxis_title is None:
            xaxis_title = XAxisAPI(name="Energy", unit="eV")
        if yaxis_title is None:
            yaxis_title = YAxisAPI(name="Intensity", unit="a.u.")
        if residual_title is None:
            residual_title = ResidualAPI(name="Residual", unit="a.u.")
        if metric_title is None:
            metric_title = MetricAPI(
                name_0="Metrics",
                unit_0="a.u.",
                name_1="Metrics",
                unit_1="a.u.",
            )
        if run_title is None:
            run_title = RunAPI(name="Run", unit="#")
        if legend is None:
            legend = LegendAPI(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="right",
                x=1,
            )
        if font is None:
            font = FontAPI(family="Open Sans, monospace", size=12, color="black")
        if color is None:
            color = ColorAPI()
        if grid is None:
            grid = GridAPI()

        self.args_plot = PlotAPI(
            x=self.x_column,
            y=self.y_column,
            title=title,
            xaxis_title=xaxis_title,
            yaxis_title=yaxis_title,
            residual_title=residual_title,
            metric_title=metric_title,
            run_title=run_title,
            legend_title=legend_title,
            show_legend=show_legend,
            legend=legend,
            font=font,
            minor_ticks=minor_ticks,
            color=color,
            grid=grid,
            size=size,
        )
        self.export_args_df = FnameAPI(fname=fname, folder=folder, suffix="csv")
        self.export_args_out = FnameAPI(fname=fname, folder=folder, suffix="lock")

        self.settings_solver_models: SolverModelsAPI = SolverModelsAPI()
        self.pre_statistic: dict[str, Any] = {}

    @property
    def pre_process(self) -> None:
        """Pre-processing class."""
        self.df, _pre_statistic = PreProcessing(
            df=self.df,
            args=self.args_pre.model_dump(),
        )()
        self.pre_statistic = _pre_statistic["data_statistic"]
        self.df_pre = self.df.copy()

    @property
    def return_pre_statistic(self) -> dict[str, Any]:
        """Return the pre-processing statistic."""
        return self.pre_statistic

    @property
    def return_df_org(self) -> pd.DataFrame:
        """Return the original dataframe."""
        return self.df_org

    @property
    def return_df_pre(self) -> pd.DataFrame | None:
        """Return the pre-processed dataframe."""
        return self.df_pre

    @property
    def return_df(self) -> pd.DataFrame:
        """Return the dataframe."""
        return self.df

    @property
    def return_df_fit(self) -> pd.DataFrame:
        """Return the fit dataframe."""
        return self.df_fit

    @property
    def export_df_act(self) -> None:
        """Export the dataframe."""
        self.export_args_df.prefix = "act"
        self.export_df(df=self.df, args=self.export_args_df)

    @property
    def export_df_fit(self) -> None:
        """Export the dataframe."""
        self.export_args_df.prefix = "fit"
        self.export_df(df=self.df_fit, args=self.export_args_df)

    @property
    def export_df_org(self) -> None:
        """Export the dataframe."""
        self.export_args_df.prefix = "org"
        self.export_df(df=self.df_org, args=self.export_args_df)

    @property
    def export_df_pre(self) -> None:
        """Export the dataframe."""
        if self.df_pre.empty is False:
            self.export_args_df.prefix = "pre"
            self.export_df(df=self.df_pre, args=self.export_args_df)

    @property
    def export_df_metric(self) -> None:
        """Export the dataframe."""
        if self.df_metric.empty is False:
            self.export_args_df.prefix = "metric"
            self.export_df(df=self.df_metric, args=self.export_args_df)

    @property
    def export_df_peaks(self) -> None:
        """Export the dataframe."""
        if self.df_peaks.empty is False:
            self.export_args_df.prefix = "peaks"
            self.export_df(df=self.df_peaks, args=self.export_args_df)

    @property
    def plot_original_df(self) -> None:
        """Plot the original spectra."""
        self.plot_dataframe(args_plot=self.args_plot, df=self.df_org)

    @property
    def plot_current_df(self) -> None:
        """Plot the current spectra."""
        self.plot_dataframe(args_plot=self.args_plot, df=self.df)

    @property
    def plot_preprocessed_df(self) -> None:
        """Plot the current processed spectra."""
        self.plot_2dataframes(
            args_plot=self.args_plot,
            df_1=self.df_pre,
            df_2=self.df_org,
        )

    def plot_fit_df(self) -> None:
        """Plot the fit."""
        if self.global_ == 1:
            self.plot_global_fit(args_plot=self.args_plot, df=self.df_fit)
        else:
            self.plot_2dataframes(args_plot=self.args_plot, df_1=self.df_fit)

    def plot_current_metric(
        self,
        bar_criteria: str | list[str] | None = None,
        line_criteria: str | list[str] | None = None,
    ) -> None:
        """Plot the current metric.

        Args:
            bar_criteria (Optional[Union[str, List[str]]], optional): Criteria for the
                    bar plot. Defaults to None.
            line_criteria (Optional[Union[str, List[str]]], optional): Criteria for
                    the line plot. Defaults to None.

        """
        if bar_criteria is None:
            bar_criteria = [
                "akaike_information",
                "bayesian_information",
            ]

        if line_criteria is None:
            line_criteria = [
                "mean_squared_error",
            ]

        self.plot_metric(
            args_plot=self.args_plot,
            df_metric=self.df_metric,
            bar_criteria=bar_criteria,
            line_criteria=line_criteria,
        )

    @property
    def generate_report(self) -> None:
        """Generate the SpectraFit report of the final fit."""
        self.export_report(
            report=ExportReport(
                description=self.args_desc,
                initial_model=self.initial_model,
                pre_processing=self.args_pre,
                settings_solver_models=self.settings_solver_models,
                fname=self.export_args_out,
                args_out=self.args,
                df_org=self.df_org,
                df_pre=self.df_pre,
                df_fit=self.df_fit,
            )(),
            args=self.export_args_out,
        )

    def solver_model(
        self,
        initial_model: list[dict[str, dict[str, dict[str, Any]]]],
        *,
        show_plot: bool = True,
        show_metric: bool = True,
        show_df: bool = False,
        show_peaks: bool = False,
        conf_interval: bool | dict[str, Any] = False,
        bar_criteria: str | list[str] | None = None,
        line_criteria: str | list[str] | None = None,
        solver_settings: dict[str, Any] | None = None,
    ) -> None:
        """Solves the fit problem based on the proposed model.

        Args:
            initial_model (List[Dict[str, Dict[str, Dict[str, Any]]]]): List of
                 dictionary with the initial model and its fitting parameters and
                 options for the components.
            show_plot (bool, optional): Show current fit results as plot.
                 Defaults to True.
            show_metric (bool, optional): Show the metric of the fit. Defaults to True.
            show_df (bool, optional): Show current fit results as dataframe. Defaults
                 to False.
            show_peaks (bool, optional): Show the peaks of fit. Defaults to False.
            conf_interval (Union[bool,Dict[str, Any]], optional): Bool or dictionary for
                 the parameter with the parameter for calculating the confidence
                 interval. Using `conf_interval=False` turns of the calculation of
                 the confidence interval and accelerate its. Defaults to False.
            bar_criteria (Optional[Union[str, List[str]]], optional): Criteria for the
                bar plot. It is recommended to use attributes from `goodness of fit`
                module. Defaults to None.
            line_criteria (Optional[Union[str, List[str]]], optional): Criteria for
                the line plot. It is recommended to use attributes from
                `regression metric` module. Defaults to None.
            solver_settings (Optional[Dict[str, Any]], optional): Settings for
                the solver models, which is split into settings for `minimizer` and
                `optimizer`.  Defaults to None.

        !!! info: "About criteria"

            The criteria for the bar and line plot are defined as a list of strings.
            The supported keywords are defined by the built-in metrics for
            `goodness of fit` and `regression` and can be checked in [documentation](
                https://anselmoo.github.io/spectrafit/doc/statistics/
            ).

        """
        self.initial_model = initial_model

        if isinstance(conf_interval, bool):
            conf_interval = (
                ConfIntervalAPI().model_dump() if conf_interval is True else False
            )
        elif isinstance(conf_interval, dict):
            conf_interval = ConfIntervalAPI(**conf_interval).dict(exclude_none=True)

        if solver_settings is not None and isinstance(solver_settings, dict):
            self.settings_solver_models = SolverModelsAPI(**solver_settings)

        self.df_fit, self.args = PostProcessing(
            self.df,
            {
                "global_": self.global_,
                "conf_interval": conf_interval,
            },
            *SolverModels(
                df=self.df,
                args={
                    "global_": self.global_,
                    "column": list(self.df.columns),
                    "autopeak": self.autopeak,
                    **list2dict(peak_list=self.initial_model),
                    **self.settings_solver_models.model_dump(),
                },
            )(),
        )()
        self.update_metric()
        self.update_peaks()
        if show_plot:
            self.plot_fit_df()

        if show_metric:
            self.plot_current_metric(
                bar_criteria=bar_criteria,
                line_criteria=line_criteria,
            )

        if show_df:
            self.interactive_display(df=self.df_fit)

        if show_peaks:
            self.interactive_display(df=self.df_peaks)

    def update_peaks(self) -> None:
        """Update the peaks dataframe as multi-column dataframe.

        The multi-column dataframe is used for the interactive display of the
        peaks with initial, current (model), and best fit values.
        """
        tuples = []
        _list = []
        for key_1, _dict in self.args["fit_insights"]["variables"].items():
            tuples.extend([(key_1, key_2) for key_2, val in _dict.items()])
            _list.extend([val for _, val in _dict.items()])

        self.df_peaks = pd.concat(
            [
                self.df_peaks,
                pd.DataFrame(
                    pd.Series(
                        _list,
                        index=pd.MultiIndex.from_tuples(
                            tuples,
                            names=["component", "parameter"],
                        ),
                    ),
                ).T,
            ],
            ignore_index=True,
        )

    def update_metric(self) -> None:
        """Update the metric dataframe."""
        self.df_metric = pd.concat(
            [self.df_metric, SolverResults(self.args).get_current_metric],
            ignore_index=True,
        )

    def display_fit_df(self, mode: str | None = "regular") -> None:
        """Display the fit dataframe.

        Args:
            mode (str, optional): Display mode. Defaults to "regular".

        """
        self.df_display(df=self.df_fit, mode=mode)

    def display_preprocessed_df(self, mode: str | None = "regular") -> None:
        """Display the preprocessed dataframe.

        Args:
            mode (str, optional): Display mode. Defaults to "regular".

        """
        self.df_display(df=self.df_pre, mode=mode)

    def display_original_df(self, mode: str | None = "regular") -> None:
        """Display the original dataframe.

        Args:
            mode (str, optional): Display mode. Defaults to "regular".

        """
        self.df_display(df=self.df_org, mode=mode)

    def display_current_df(self, mode: str | None = "regular") -> None:
        """Display the current dataframe.

        Args:
            mode (str, optional): Display mode. Defaults to "regular".

        """
        self.df_display(df=self.df, mode=mode)



"""Physical constants for use across spectrafit models."""

from __future__ import annotations

from pydantic import BaseModel
from pydantic import Field


class MoessbauerConstants(BaseModel):
    """Physical constants specific to Mössbauer spectroscopy."""

    # General nuclear physics constants
    nuclear_magneton: float = Field(
        default=5.0508e-27,
        description="Nuclear magneton in J/T",
    )

    # 57Fe specific constants
    gamma_57fe: float = Field(
        default=8.67e-9,
        description="Natural linewidth of 57Fe in eV",
    )
    g_factor_57fe: float = Field(
        default=0.18,
        description="g-factor for 57Fe excited state",
    )
    quadrupole_moment_57fe: float = Field(
        default=0.16e-28,
        description="Quadrupole moment for 57Fe in m²",
    )
    conversion_mm_s_to_ev: float = Field(
        default=4.8e-8,
        description="Conversion from mm/s to eV for 57Fe (approx)",
    )
    ev_to_mm_s: float = Field(
        default=1 / 4.8e-8,
        description="Convert energy in eV to velocity in mm/s",
    )

    # Thresholds for model calculations
    min_efg_threshold: float = Field(
        default=1e20,
        description="Minimum EFG value for octet splitting in V/m²",
    )
    min_field_threshold: float = Field(
        default=1.0,
        description="Minimum magnetic field value for octet splitting in T",
    )
    min_variance_threshold: float = Field(
        default=1e-6,
        description="Minimum variance for fitting in eV²",
    )


# Create a singleton instance for easy import
moessbauer_constants = MoessbauerConstants()



"""Minimization models for curve fitting."""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from math import log
from math import pi
from math import sqrt
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar

import numpy as np

from lmfit import Minimizer
from lmfit import Parameters
from scipy.signal import find_peaks
from scipy.stats import hmean

from spectrafit.api.models_model import DistributionModelAPI
from spectrafit.api.tools_model import AutopeakAPI
from spectrafit.api.tools_model import GlobalFittingAPI
from spectrafit.api.tools_model import SolverModelsAPI
from spectrafit.models.moessbauer import moessbauer_doublet as _moessbauer_doublet
from spectrafit.models.moessbauer import moessbauer_octet as _moessbauer_octet
from spectrafit.models.moessbauer import moessbauer_sextet as _moessbauer_sextet
from spectrafit.models.moessbauer import moessbauer_singlet as _moessbauer_singlet
from spectrafit.models.regular import atan_step as _atan
from spectrafit.models.regular import cgaussian as _cgaussian
from spectrafit.models.regular import clorentzian as _clorentzian
from spectrafit.models.regular import constant as _constant
from spectrafit.models.regular import cvoigt as _cvoigt
from spectrafit.models.regular import erf_step as _erf
from spectrafit.models.regular import exponential as _exponential
from spectrafit.models.regular import gaussian as _gaussian
from spectrafit.models.regular import heaviside as _heaviside
from spectrafit.models.regular import linear as _linear
from spectrafit.models.regular import log_step as _log
from spectrafit.models.regular import lorentzian as _lorentzian
from spectrafit.models.regular import orcagaussian as _orcagaussian
from spectrafit.models.regular import pearson1 as _pearson1
from spectrafit.models.regular import pearson2 as _pearson2
from spectrafit.models.regular import pearson3 as _pearson3
from spectrafit.models.regular import pearson4 as _pearson4
from spectrafit.models.regular import polynom2 as _polynom2
from spectrafit.models.regular import polynom3 as _polynom3
from spectrafit.models.regular import power as _power
from spectrafit.models.regular import pseudovoigt as _pseudovoigt
from spectrafit.models.regular import voigt as _voigt


if TYPE_CHECKING:
    import pandas as pd

    from numpy.typing import NDArray

# Constants for global fitting modes
GLOBAL_NONE = 0  # No global fitting
GLOBAL_STANDARD = 1  # Standard global fitting
GLOBAL_WITH_PRE = 2  # Global fitting with pre-processing


class DistributionModels:
    """Distribution models for the fit.

    !!! note "About distribution models"

        `DistributionModels` are wrapper functions for the distribution models. The
        overall goal is to extract from the best parameters the single contributions in
        the model. The superposition of the single contributions is the final model.

    !!! note "About the cumulative distribution"

        The cumulative distribution is the sum of the single contributions. The
        cumulative distribution is the model that is fitted to the data. In contrast to
        the single contributions, the cumulative distribution is not normalized and
        therefore the amplitude of the single contributions is not directly comparable
        to the amplitude of the cumulative distribution. Also, the cumulative
        distributions are consequently using the `fwhm` parameter instead of the
        `sigma` parameter.
    """

    @staticmethod
    def gaussian(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhmg: float = 1.0,
    ) -> NDArray[np.float64]:
        """Return the wrapper function for the Gaussian model from regular_models."""
        return _gaussian(x=x, amplitude=amplitude, center=center, fwhmg=fwhmg)

    @staticmethod
    def orcagaussian(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        width: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Gaussian distribution as implemented in the ORCA program.

        $$
        {\displaystyle g(x)= A \cdot \exp
        (  -{\frac {(x-x_0)^{2}}{2 \cdot width^{2}}} ) }
        $$

        Unlike the standard gaussian function, this implementation uses the width parameter
        directly without conversion to sigma, which is consistent with the ORCA quantum
        chemistry program's approach[^1].

        [^1]: https://www.faccts.de/docs/orca/6.0/manual/contents/detailed/utilities.html

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Gaussian distribution.
                 Defaults to 1.0.
            center (float, optional): Center of the Gaussian distribution.
                 Defaults to 0.0.
            width (float, optional): Width parameter of the Gaussian distribution as used
                 in the ORCA program. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Gaussian distribution of `x` given.

        """
        return _orcagaussian(x=x, amplitude=amplitude, center=center, width=width)

    @staticmethod
    def lorentzian(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhml: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Lorentzian distribution.

        $$
        f(x;x_{0},\gamma )={\frac  {1}{\pi \gamma
        [ 1+ ( {\frac  {x-x_{0}}{\gamma }})^{2} ]
        }} ={1 \over \pi \gamma } [ {\gamma ^{2} \over (x-x_{0})^{2}+\gamma ^{2}} ]
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Lorentzian distribution.
                Defaults to 1.0.
            center (float, optional): Center of the Lorentzian distribution. Defaults to
                0.0.
            fwhml (float, optional): Full width at half maximum (FWHM) of the Lorentzian
                distribution. Defaults to 1.0.

        Returns:
            Union[NDArray[np.float64], float]: Lorentzian distribution of `x` given.

        """
        return _lorentzian(x=x, amplitude=amplitude, center=center, fwhml=fwhml)

    @staticmethod
    def voigt(
        x: NDArray[np.float64],
        center: float = 0.0,
        fwhmv: float = 1.0,
        gamma: float | None = None,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Voigt distribution.

        $$
        {\displaystyle V(x;\sigma ,\gamma )\equiv
        \int_{-\infty }^{\infty }G(x';\sigma )
        L(x-x';\gamma )\,dx'}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            center (float, optional): Center of the Voigt distribution. Defaults to 0.0.
            fwhmv (float, optional): Full width at half maximum (FWHM) of the Lorentzian
                distribution. Defaults to 1.0.
            gamma (float, optional): Scaling factor of the complex part of the
                [Faddeeva Function](https://en.wikipedia.org/wiki/Faddeeva_function).
                Defaults to None.

        Returns:
            NDArray[np.float64]: Voigt distribution of `x` given.

        """
        return _voigt(x=x, center=center, fwhmv=fwhmv, gamma=gamma)

    @staticmethod
    def pseudovoigt(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhmg: float = 1.0,
        fwhml: float = 1.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional Pseudo-Voigt distribution.

        !!! note "See also:"

            J. Appl. Cryst. (2000). 33, 1311-1316
            https://doi.org/10.1107/S0021889800010219

        Args:
            x (NDArray[np.float64]):  `x`-values of the data.
            amplitude (float, optional): Amplitude of the Pseudo-Voigt distribution.
                Defaults to 1.0.
            center (float, optional): Center of the Pseudo-Voigt distribution.
                Defaults to 0.0.
            fwhmg (float, optional): Full width half maximum of the Gaussian
                distribution in the Pseudo-Voigt distribution. Defaults to 1.0.
            fwhml (float, optional): Full width half maximum of the Lorentzian
                distribution in the Pseudo-Voigt distribution. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Pseudo-Voigt distribution of `x` given.

        """
        return _pseudovoigt(
            x=x,
            amplitude=amplitude,
            center=center,
            fwhmg=fwhmg,
            fwhml=fwhml,
        )

    @staticmethod
    def exponential(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        decay: float = 1.0,
        intercept: float = 0.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional exponential decay.

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the exponential function.
                 Defaults to 1.0.
            decay (float, optional): Decay of the exponential function. Defaults to 1.0.
            intercept (float, optional): Intercept of the exponential function.
                 Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Exponential decay of `x` given.

        """
        return _exponential(x=x, amplitude=amplitude, decay=decay, intercept=intercept)

    @staticmethod
    def power(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        exponent: float = 1.0,
        intercept: float = 0.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional power function.

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the power function. Defaults to
                1.0.
            exponent (float, optional): Exponent of the power function. Defaults to 1.0.
            intercept (float, optional): Intercept of the power function. Defaults to
                0.0.

        Returns:
            NDArray[np.float64]: power function of `x` given.

        """
        return _power(x=x, amplitude=amplitude, exponent=exponent, intercept=intercept)

    @staticmethod
    def linear(
        x: NDArray[np.float64],
        slope: float = 1.0,
        intercept: float = 0.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional linear function.

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            slope (float, optional): Slope of the linear function. Defaults to 1.0.
            intercept (float, optional): Intercept of the linear function.
                 Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Linear function of `x` given.

        """
        return _linear(x=x, slope=slope, intercept=intercept)

    @staticmethod
    def constant(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional constant value.

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the constant. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Constant value of `x` given.

        """
        return _constant(x=x, amplitude=amplitude)

    @staticmethod
    def erf(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional error function.

        $$
        f(x) = \frac{2}{\sqrt{\pi}} \int_{0}^{x} e^{-t^2} dt
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the error function.
                    Defaults to 1.0.
            center (float, optional): Center of the error function. Defaults to 0.0.
            sigma (float, optional): Sigma of the error function. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Error function of `x` given.

        """
        return _erf(x=x, amplitude=amplitude, center=center, sigma=sigma)

    @staticmethod
    def heaviside(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Heaviside step function.

        $$
        f(x) = \begin{cases}
        0 & x < 0 \\
        0.5 & x = 0 \\
        1 & x > 0
        \end{cases}
        $$
        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Heaviside step function.
                    Defaults to 1.0.
            center (float, optional): Center of the Heaviside step function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the Heaviside step function.
                 Defaults to 1.0.


        Returns:
            NDArray[np.float64]: Heaviside step function of `x` given.

        """
        return _heaviside(x=x, amplitude=amplitude, center=center, sigma=sigma)

    @staticmethod
    def atan(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional arctan step function.

        $$
        f(x) = \frac{1}{\pi} \arctan(\frac{x - c}{s})
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the arctan step function.
                    Defaults to 1.0.
            center (float, optional): Center of the arctan step function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the arctan step function.
                 Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Arctan step function of `x` given.

        """
        return _atan(x=x, amplitude=amplitude, center=center, sigma=sigma)

    @staticmethod
    def log(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional logarithmic step function.

        $$
        f(x) = \frac{1}{1 + e^{-\frac{x - c}{s}}}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the logarithmic step function.
                    Defaults to 1.0.
            center (float, optional): Center of the logarithmic step function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the logarithmic step function.
                 Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Logarithmic step function of `x` given.

        """
        return _log(x=x, amplitude=amplitude, center=center, sigma=sigma)

    @staticmethod
    def cgaussian(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhmg: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional cumulative Gaussian function.

        $$
        f(x) = \frac{1}{2} \left[1 + erf\left(\frac{x - c}{s \sqrt{2}}\right)\right]
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Gaussian function. Defaults to
                1.0.
            center (float, optional): Center of the Gaussian function. Defaults to 0.0.
            fwhmg (float, optional): Full width at half maximum of the Gaussian
                 function. Defaults to 1.0.


        Returns:
            NDArray[np.float64]: Cumulative Gaussian function of `x` given.

        """
        return _cgaussian(x=x, amplitude=amplitude, center=center, fwhmg=fwhmg)

    @staticmethod
    def clorentzian(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhml: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional cumulative Lorentzian function.

        $$
        f(x) = \frac{1}{\pi} \arctan\left(\frac{x - c}{s}\right) + \frac{1}{2}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Lorentzian function.
                    Defaults to 1.0.
            center (float, optional): Center of the Lorentzian function.
                 Defaults to 0.0.
            fwhml (float, optional): Full width at half maximum of the Lorentzian
                function. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Cumulative Lorentzian function of `x` given.

        """
        return _clorentzian(x=x, amplitude=amplitude, center=center, fwhml=fwhml)

    @staticmethod
    def cvoigt(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        fwhmv: float = 1.0,
        gamma: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional cumulative Voigt function.

        $$
        f(x) = \frac{1}{2} \left[1 + erf\left(\frac{x - c}{s \sqrt{2}}\right)\right]
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Voigt function. Defaults to
                1.0.
            center (float, optional): Center of the Voigt function. Defaults to 0.0.
            fwhmv (float, optional): Full width at half maximum of the Voigt function.
                Defaults to 1.0.
            gamma (float, optional): Gamma of the Voigt function. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Cumulative Voigt function of `x` given.

        """
        return _cvoigt(
            x=x,
            amplitude=amplitude,
            center=center,
            fwhmv=fwhmv,
            gamma=gamma,
        )

    @staticmethod
    def polynom2(
        x: NDArray[np.float64],
        coefficient0: float = 1.0,
        coefficient1: float = 1.0,
        coefficient2: float = 1.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional second order polynomial function.

        $$
        f(x) = c_2 x^2 + c_1 x + c_0
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data
            coefficient0 (float, optional): Zeroth coefficient of the
                 polynomial function. Defaults to 1.0.
            coefficient1 (float, optional): First coefficient of the
                 polynomial function. Defaults to 1.0.
            coefficient2 (float, optional): Second coefficient of the
                 polynomial function. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Third order polynomial function of `x`

        """
        return _polynom2(
            x=x,
            coefficient0=coefficient0,
            coefficient1=coefficient1,
            coefficient2=coefficient2,
        )

    @staticmethod
    def polynom3(
        x: NDArray[np.float64],
        coefficient0: float = 1.0,
        coefficient1: float = 1.0,
        coefficient2: float = 1.0,
        coefficient3: float = 1.0,
    ) -> NDArray[np.float64]:
        """Return a 1-dimensional third order polynomial function.

        $$
        f(x) = c_3 x^3 + c_2 x^2 + c_1 x + c_0
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data
            coefficient0 (float, optional): Zeroth coefficient of the
                 polynomial function. Defaults to 1.0.
            coefficient1 (float, optional): First coefficient of the
                 polynomial function. Defaults to 1.0.
            coefficient2 (float, optional): Second coefficient of the
                 polynomial function. Defaults to 1.0.
            coefficient3 (float, optional): Third coefficient of the
                 polynomial function. Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Third order polynomial function of `x`

        """
        return _polynom3(
            x=x,
            coefficient0=coefficient0,
            coefficient1=coefficient1,
            coefficient2=coefficient2,
            coefficient3=coefficient3,
        )

    @staticmethod
    def pearson1(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
        exponent: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Pearson type I distribution.

        $$
        f(x) = \frac{A}{\sigma \sqrt{2 \pi}} \left[1 + \frac{(x - c)^2}{\sigma^2}
        \right]^{-\frac{1}{\nu}}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Pearson type I function.
                    Defaults to 1.0.
            center (float, optional): Center of the Pearson type I function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the Pearson type I function.
                 Defaults to 1.0.
            exponent (float, optional): Exponent of the Pearson type I function.
                 Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Pearson type I function of `x` given.

        """
        return _pearson1(
            x=x,
            amplitude=amplitude,
            center=center,
            sigma=sigma,
            exponent=exponent,
        )

    @staticmethod
    def pearson2(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
        exponent: float = 1.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Pearson type II distribution.

        $$
        f(x) = \frac{A}{\sigma \sqrt{2 \pi}} \left[1 + \frac{(x - c)^2}{2 \sigma^2}
        \right]^{-\nu}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Pearson type II function.
                    Defaults to 1.0.
            center (float, optional): Center of the Pearson type II function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the Pearson type II function.
                 Defaults to 1.0.
            exponent (float, optional): Exponent of the Pearson type II function.
                 Defaults to 1.0.

        Returns:
            NDArray[np.float64]: Pearson type II function of `x` given.

        """
        return _pearson2(
            x=x,
            amplitude=amplitude,
            center=center,
            sigma=sigma,
            exponent=exponent,
        )

    @staticmethod
    def pearson3(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
        exponent: float = 1.0,
        skewness: float = 0.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Pearson type III distribution.

        $$
        f(x) = \frac{A}{\sigma \sqrt{2 \pi}} \left[1 + \frac{(x - c)^2}{2 \sigma^2}
        \right]^{-\nu} \left[1 + \frac{\gamma}{\nu}
        \frac{x - c}{\sigma} \right]^{-\nu - 1}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Pearson type III function.
                    Defaults to 1.0.
            center (float, optional): Center of the Pearson type III function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the Pearson type III function.
                 Defaults to 1.0.
            exponent (float, optional): Exponent of the Pearson type III function.
                 Defaults to 1.0.
            skewness (float, optional): Skewness of the Pearson type III function.
                 Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Pearson type III function of `x` given.

        """
        return _pearson3(
            x=x,
            amplitude=amplitude,
            center=center,
            sigma=sigma,
            exponent=exponent,
            skewness=skewness,
        )

    @staticmethod
    def pearson4(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        center: float = 0.0,
        sigma: float = 1.0,
        exponent: float = 1.0,
        skewness: float = 0.0,
        kurtosis: float = 0.0,
    ) -> NDArray[np.float64]:
        r"""Return a 1-dimensional Pearson type IV distribution.

        $$
        f(x) = \frac{A}{\sigma \sqrt{2 \pi}} \left[1 + \frac{(x - c)^2}{2 \sigma^2}
        \right]^{-\nu} \left[1 + \frac{\gamma}{\nu}
        \frac{x - c}{\sigma} \right]^{-\nu - 1}
        \left[1 + \frac{\delta}{\nu}
        \left(\frac{x - c}{\sigma}\right)^2 \right]^{-\nu - 1/2}
        $$

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            amplitude (float, optional): Amplitude of the Pearson type IV function.
                    Defaults to 1.0.
            center (float, optional): Center of the Pearson type IV function.
                 Defaults to 0.0.
            sigma (float, optional): Sigma of the Pearson type IV function.
                 Defaults to 1.0.
            exponent (float, optional): Exponent of the Pearson type IV function.
                 Defaults to 1.0.
            skewness (float, optional): Skewness of the Pearson type IV function.
                 Defaults to 0.0.
            kurtosis (float, optional): Kurtosis of the Pearson type IV function.
                 Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Pearson type IV function of `x` given.

        """
        return _pearson4(
            x=x,
            amplitude=amplitude,
            center=center,
            sigma=sigma,
            exponent=exponent,
            skewness=skewness,
            kurtosis=kurtosis,
        )

    @staticmethod
    def moessbauersinglet(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        isomershift: float = 0.0,
        fwhml: float = 0.25,
        center: float = 0.0,
        background: float = 0.0,
    ) -> NDArray[np.float64]:
        """Compute a Mössbauer singlet via Lorentzian plus background.

        Args:
            x (NDArray[np.float64]): Velocity array (mm/s).
            amplitude (float, optional): Peak amplitude. Defaults to 1.0.
            isomershift (float, optional): Isomer shift (mm/s). Defaults to 0.0.
            fwhml (float, optional): Lorentzian full-width at half-maximum. Defaults to 0.25.
            center (float, optional): Global spectrum offset in mm/s. Defaults to 0.0.
            background (float, optional): Constant background level. Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Computed transmission intensity array.

        """
        # Call the implementation from moessbauer_models.py
        return _moessbauer_singlet(
            x=x,
            amplitude=amplitude,
            isomer_shift=isomershift,
            fwhml=fwhml,
            center=center,
            background=background,
        )

    @staticmethod
    def moessbauerdoublet(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        isomershift: float = 0.4,
        fwhml: float = 0.25,
        quadrupolesplitting: float = 0.8,
        center: float = 0.0,
        background: float = 0.0,
    ) -> NDArray[np.float64]:
        """Compute a Mössbauer doublet via two Lorentzians plus background.

        Args:
            x (NDArray[np.float64]): Velocity or energy array (mm/s).
            amplitude (float, optional): Peak amplitude. Defaults to 1.0.
            isomershift (float, optional): Isomer shift (mm/s). Defaults to 0.4.
            fwhml (float, optional): Lorentzian full-width at half-maximum. Defaults to 0.25.
            quadrupolesplitting (float, optional): Quadrupole splitting (mm/s). Defaults to 0.8.
            center (float, optional): Global spectrum offset in mm/s. Defaults to 0.0.
            background (float, optional): Constant background level. Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Computed intensity array.

        """
        # Call the implementation from moessbauer_models.py
        return _moessbauer_doublet(
            x=x,
            amplitude=amplitude,
            isomer_shift=isomershift,
            fwhml=fwhml,
            quadrupole_splitting=quadrupolesplitting,
            center=center,
            background=background,
        )

    @staticmethod
    def moessbauersextet(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        isomershift: float = 0.0,
        fwhml: float = 0.25,
        magneticfield: float = 33.0,
        quadrupoleshift: float = 0.0,
        center: float = 0.0,
        background: float = 0.0,
        anglethetaphi: dict[str, float] | None = None,
    ) -> NDArray[np.float64]:
        """Compute a Mössbauer sextet via six Lorentzians plus background.

        Args:
            x (NDArray[np.float64]): Velocity or energy array (mm/s).
            amplitude (float, optional): Peak amplitude. Defaults to 1.0.
            isomershift (float, optional): Isomer shift (mm/s). Defaults to 0.0.
            fwhml (float, optional): Lorentzian full-width at half-maximum. Defaults to 0.25.
            magneticfield (float, optional): Hyperfine field in Tesla. Defaults to 33.0.
            quadrupoleshift (float, optional): First-order quadrupole shift in mm/s.
                Defaults to 0.0.
            center (float, optional): Global spectrum offset in mm/s. Defaults to 0.0.
            background (float, optional): Constant background level. Defaults to 0.0.
            anglethetaphi (Dict[str, float], optional): Orientation angles. Defaults to None.

        Returns:
            NDArray[np.float64]: Computed intensity array.
        """
        return _moessbauer_sextet(
            x=x,
            amplitude=amplitude,
            isomer_shift=isomershift,
            fwhml=fwhml,
            magnetic_field=magneticfield,
            quadrupole_shift=quadrupoleshift,
            center=center,
            angle_theta_phi=anglethetaphi,
            background=background,
        )

    @staticmethod
    def moessbaueroctet(
        x: NDArray[np.float64],
        amplitude: float = 1.0,
        isomershift: float = 0.0,
        fwhml: float = 0.25,
        magneticfield: float = 33.0,
        quadrupoleshift: float = 0.0,
        center: float = 0.0,
        efg_vzz: float = 1e21,
        efg_eta: float = 0.0,
        anglethetaphi: dict[str, float] | None = None,
        temperature: float = 300.0,
        sodshift: float = 0.0,
        sitefraction: float = 1.0,
        background: float = 0.0,
    ) -> NDArray[np.float64]:
        """Compute a Mössbauer octet via eight Lorentzians plus background.

        Used for materials with a mixture of magnetic and quadrupole interactions.

        Args:
            x (NDArray[np.float64]): Velocity or energy array (mm/s).
            amplitude (float, optional): Peak amplitude. Defaults to 1.0.
            isomershift (float, optional): Isomer shift (mm/s). Defaults to 0.0.
            fwhml (float, optional): Lorentzian full-width at half-maximum. Defaults to 0.25.
            magneticfield (float, optional): Hyperfine field in Tesla. Defaults to 33.0.
            quadrupoleshift (float, optional): First-order quadrupole shift in mm/s.
                Defaults to 0.0.
            center (float, optional): Global spectrum offset in mm/s. Defaults to 0.0.
            efg_vzz (float, optional): Principal component of EFG tensor. Defaults to 1e21.
            efg_eta (float, optional): EFG asymmetry parameter. Defaults to 0.0.
            anglethetaphi (Dict[str, float], optional): Orientation angles. Defaults to None.
            temperature (float, optional): Temperature in K. Defaults to 300.0.
            sodshift (float, optional): Second-order Doppler shift in mm/s. Defaults to 0.0.
            sitefraction (float, optional): Site fraction. Defaults to 1.0.
            background (float, optional): Constant background level. Defaults to 0.0.

        Returns:
            NDArray[np.float64]: Computed intensity array.
        """
        # Call the implementation from moessbauer_models.py
        return _moessbauer_octet(
            x=x,
            amplitude=amplitude,
            isomer_shift=isomershift,
            fwhml=fwhml,
            magnetic_field=magneticfield,
            quadrupole_shift=quadrupoleshift,
            center=center,
            efg_vzz=efg_vzz,
            efg_eta=efg_eta,
            angle_theta_phi=anglethetaphi,
            temperature=temperature,
            sod_shift=sodshift,
            site_fraction=sitefraction,
            background=background,
        )


@dataclass(frozen=True)
class ReferenceKeys:
    """Reference keys for model fitting and peak detection."""

    __models__: ClassVar[list[str]] = list(
        DistributionModelAPI.model_json_schema()["properties"].keys(),
    )

    __automodels__: ClassVar[list[str]] = [
        "gaussian",
        "orcagaussian",
        "lorentzian",
        "voigt",
        "pseudovoigt",
    ]

    # Mössbauer models
    __moessbauer_models__: ClassVar[list[str]] = [
        "moessbauersinglet",
        "moessbauerdoublet",
        "moessbauersextet",
        "moessbaueroctet",
    ]

    def model_check(self, model: str) -> None:
        """Check if model is available.

        Args:
            model (str): Model name.

        Raises:
            NotImplementedError: If the model is not implemented.

        """
        model_prefix = model.split("_")[0]

        # Check in main models list
        if (
            model_prefix not in self.__models__
            and model_prefix not in self.__moessbauer_models__
        ):
            msg = f"{model} is not supported!"
            raise NotImplementedError(msg)

    def automodel_check(self, model: str) -> None:
        """Check if model is available.

        Args:
            model (str): Auto Model name (gaussian, orcagaussian,
                lorentzian, voigt, or pseudovoigt).

        Raises:
            KeyError: If the model is not supported.

        """
        if model not in self.__automodels__:
            msg = f"{model} is not supported for auto detection! Use one of {self.__automodels__}"
            raise KeyError(msg)

    def detection_check(self, args: dict[str, Any]) -> None:
        """Check if detection is available.

        Args:
            args (Dict[str, Any]): The input file arguments as a dictionary with
                 additional information beyond the command line arguments.

        Raises:
            KeyError: If the key is not parameter of the `scipy.signal.find_peaks`
                function. This will be checked via `pydantic` in `spectrafit.api`.

        """
        AutopeakAPI(**args)


class AutoPeakDetection:
    """Automatic detection of peaks in a spectrum."""

    def __init__(
        self,
        x: NDArray[np.float64],
        data: NDArray[np.float64],
        args: dict[str, Any],
    ) -> None:
        """Initialize the AutoPeakDetection class.

        Args:
            x (NDArray[np.float64]): `x`-values of the data.
            data (NDArray[np.float64]): `y`-values of the data as 1d-array.
            args (Dict[str, Any]): The input file arguments as a dictionary with
                 additional information beyond the command line arguments.

        """
        self.x = x
        self.data = data
        self._args = args["autopeak"]

    @staticmethod
    def check_key_exists(
        key: str,
        args: dict[str, Any],
        value: float | tuple[Any, Any],
    ) -> Any:
        """Check if a key exists in a dictionary.

        Please check for the reference key also [scipy.signal.find_peaks][1].

        [1]:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html


        Args:
            key (str): Reference key of `scipy.signal.find_peaks`.
            args (Dict[str, Any]): Reference values of `scipy.signal.find_peaks`, if not
                 defined will be set to estimated default values.
            value (Union[float, Tuple[float,float]]): Default value for the reference
                 key.

        Returns:
            Any: The reference value for `scipy.signal.find_peaks`.

        """
        return args.get(key, value)

    @property
    def estimate_height(self) -> tuple[float, float]:
        r"""Estimate the initial height based on an inverse noise ratio of a signal.

        !!! info "About the estimation of the height"

            The lower end of the height is the inverse noise ratio of the `data`, and
            upper limit is the maximum value of the `data`. The noise ratio of the
            `data` is based on the original implementation by `SciPy`:

            ```python
            def signaltonoise(a, axis=0, ddof=0):
                a = np.asanyarray(a)
                m = a.mean(axis)
                sd = a.std(axis=axis, ddof=ddof)
                return np.where(sd == 0, 0, m / sd)
            ```

        Returns:
            Tuple[float, float]: Tuple of the inverse signal to noise ratio and
                 the maximum value of the `data`.

        """
        return 1 - self.data.mean() / self.data.std(), self.data.max()

    @property
    def estimate_threshold(self) -> tuple[float, float]:
        """Estimate the threshold value for the peak detection.

        Returns:
            Tuple[float, float]: Minimum and maximum value of the spectrum `data`,
                 respectively, `intensity`.

        """
        return self.data.min(), self.data.max()

    @property
    def estimate_distance(self) -> float:
        """Estimate the initial distance between peaks.

        Returns:
            float: Estimated distance between peaks.

        """
        min_step = np.diff(self.x).min()
        return max(min_step, 1.0)

    @property
    def estimate_prominence(self) -> tuple[float, float]:
        """Estimate the prominence of a peak.

        !!! info "About the estimation of the prominence"

            The prominence is the difference between the height of the peak and the
            bottom. To get a estimate of the prominence, the height of the peak is
            calculated by maximum value of the `data` and the bottom is calculated by
            the harmonic mean of the `data`.

        Returns:
            Tuple[float, float]: Tuple of the harmonic-mean and maximum value of `data`.

        """
        try:
            return hmean(self.data), self.data.max()
        except ValueError:
            pass
        return self.data.mean(), self.data.max()

    @property
    def estimated_width(self) -> tuple[float, float]:
        """Estimate the width of a peak.

        !!! info "About the estimation of the width"

            The width of a peak is estimated for a lower and an upper end. For the lower
            end, the minimum stepsize is used. For the upper end, the stepsize between
            the half maximum and the minimum value of the `data` is used as the width.

        Returns:
            Tuple[float, float]: Estimated width lower and uper end of the peaks.

        """
        return (
            np.diff(self.x).min(),
            np.abs(self.x[self.data.argmax()] - self.x[self.data.argmin()]) / 2,
        )

    @property
    def estimated_rel_height(self) -> float:
        """Estimate the relative height of a peak.

        !!! info "About the estimation of the relative height"

            The relative height of a peak is approximated by the difference of the
            harmonic mean value of the `data` and the minimum value of the `data`
            divided by the factor of `4`. In case of negative ratios, the value will be
            set to `Zero`.

        Returns:
            float: Estimated relative height of a peak.

        """
        try:
            rel_height = (hmean(self.data) - self.data.min()) / 4
        except ValueError:
            rel_height = (self.data.mean() - self.data.min()) / 4
        return rel_height if rel_height > 0 else 0.0

    @property
    def estimated_wlen(self) -> float:
        r"""Estimate the window length for the peak detection.

        !!! info "About the estimation of the window length"

            The window length is the length of the window for the peak detection is
            defined to be 1% of the length of the `data`, consequently the len of the
            `data` is divided by 100. In case of a window length smaller than 1, the
            window length will be set to numerical value of 1, which is defined by
            `1 + 1e-9`.

        Returns:
            float: Estimated window length is set to the numeric value of > 1.

        """
        wlen = self.data.size / 100
        return wlen if wlen > 1.0 else 1 + 1e-9

    @property
    def estimated_plateau_size(self) -> tuple[float, float]:
        """Estimate the plateau size for the peak detection.

        Returns:
            Tuple[float, float]: Estimated plateau size is set to `zero` for the lower
                 end and the maximum value of the `x` for the upper end.

        """
        return 0.0, self.x.max()

    def initialize_peak_detection(self) -> None:
        """Initialize the peak detection.

        !!! note "Initialize the peak detection"

            This method is used to initialize the peak detection. The initialization can
            be activated by setting the `initialize` attribute to `True`, which will
            automatically estimate the default parameters for the peak detection. In
            case of the `initialize` attribute is defined as dictionary, the proposed
            values are taken from the dictionary if th

        Raise:
            TypeError: If the `initialize` attribute is not of type `bool` or `dict`.
        """
        if isinstance(self._args, bool):
            self.default_values()
        elif isinstance(self._args, dict):
            ReferenceKeys().detection_check(self._args)
            self.height = self.check_key_exists(
                key="height",
                args=self._args,
                value=self.estimate_height,
            )
            self.threshold = self.check_key_exists(
                key="threshold",
                args=self._args,
                value=self.estimate_threshold,
            )
            self.distance = self.check_key_exists(
                key="distance",
                args=self._args,
                value=self.estimate_distance,
            )
            self.prominence = self.check_key_exists(
                key="prominence",
                args=self._args,
                value=self.estimate_prominence,
            )
            self.width = self.check_key_exists(
                key="width",
                args=self._args,
                value=self.estimated_width,
            )
            self.wlen = self.check_key_exists(
                key="wlen",
                args=self._args,
                value=self.estimated_wlen,
            )
            self.rel_height = self.check_key_exists(
                key="rel_height",
                args=self._args,
                value=self.estimated_rel_height,
            )
            self.plateau_size = self.check_key_exists(
                key="plateau_size",
                args=self._args,
                value=0.0,
            )
        else:
            msg = f"The type of the `args` is not supported: {type(self._args)}"
            raise TypeError(msg)

    def default_values(self) -> None:
        """Set the default values for the peak detection."""
        self.height = self.estimate_height
        self.threshold = self.estimate_threshold
        self.distance = self.estimate_distance
        self.prominence = self.estimate_prominence
        self.width = self.estimated_width
        self.wlen = self.estimated_wlen
        self.rel_height = self.estimated_rel_height
        self.plateau_size = 0

    def __autodetect__(self) -> Any:
        """Return peak positions and properties."""
        return find_peaks(
            self.data,
            height=self.height,
            threshold=self.threshold,
            distance=self.distance,
            prominence=self.prominence,
            width=self.width,
            wlen=self.wlen,
            rel_height=self.rel_height,
            plateau_size=self.plateau_size,
        )


class ModelParameters(AutoPeakDetection):
    """Class to define the model parameters."""

    def __init__(self, df: pd.DataFrame, args: dict[str, Any]) -> None:
        """Initialize the model parameters.

        Args:
            df (pd.DataFrame): DataFrame containing the input data (`x` and `data`).
            args (Dict[str, Any]):
                 Nested arguments dictionary for the model based on **one** or **two**
                 `int` keys depending if global fitting parameters, will explicit
                 defined or not.

        !!! note "About `args` for models"

            The `args` dictionary is used to define the model parameters. And the total
            nested dictionary structure is as follows:

            ```python
            args: Dict[str, Dict[int, Dict[str, Dict[str, Union[str, int, float]]]]]
            ```

        !!! info "About the fitting options"

            In general, there are two option for the fitting possible:

            1. `Classic fitting` or `local fitting`, where the parameters are defined
                for a 2D spectrum.
            2. `Global fitting`, where the parameters are defined for a 3D spectrum.
                Here, the parameters can be automatically defined for each column on the
                basis of the initial parameters or they can be completley defined by the
                user. The `global fitting` definition starts at `1` similiar to the
                peaks attributes notation.

        """
        self.col_len = df.shape[1] - 1
        self.args = args
        self.params = Parameters()
        self.x, self.data = self.df_to_numvalues(df=df, args=args)

        super().__init__(self.x, self.data, self.args)

    def df_to_numvalues(
        self,
        df: pd.DataFrame,
        args: dict[str, Any],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Transform the dataframe to numeric values of `x` and `data`.

        !!! note "About the dataframe to numeric values"

            The transformation is done by the `value` property of pandas. The dataframe
            is separated into the `x` and `data` columns and the `x` column is
            transformed to the energy values and the `data` column is transformed to
            the intensity values depending on the `args` dictionary. In terms of global
            fitting, the `data` contains the intensity values for each column.

        Args:
            df (pd.DataFrame): DataFrame containing the input data (`x` and `data`).
            args (Dict[str, Any]): The input file arguments as a dictionary with
                 additional information beyond the command line arguments.

        Returns:
            Tuple[NDArray[np.float64], NDArray[np.float64]]: Tuple of `x` and
                 `data` as numpy arrays.

        """
        if args["global_"]:
            return (
                df[args["column"][0]].to_numpy(),
                df.loc[:, df.columns != args["column"][0]].to_numpy(),
            )
        return (df[args["column"][0]].to_numpy(), df[args["column"][1]].to_numpy())

    @property
    def return_params(self) -> Parameters:
        """Return the `class` representation of the model parameters.

        Returns:
            Parameters: Model parameters class.

        """
        self.__perform__()
        return self.params

    def __str__(self) -> str:
        """Return the `string` representation of the model parameters.

        Returns:
            str: String representation of the model parameters.

        """
        self.__perform__()
        return str(self.params)

    def __perform__(self) -> None:
        """Perform the model parameter definition.

        Raises:
            KeyError: Global fitting is combination with automatic peak detection is
                 not implemented yet.

        """
        if self.args["global_"] == GLOBAL_NONE and not self.args["autopeak"]:
            self.define_parameters()
        elif self.args["global_"] == GLOBAL_STANDARD and not self.args["autopeak"]:
            self.define_parameters_global()
        elif self.args["global_"] == GLOBAL_WITH_PRE and not self.args["autopeak"]:
            self.define_parameters_global_pre()
        elif self.args["global_"] == GLOBAL_NONE:
            self.initialize_peak_detection()
            self.define_parameters_auto()
        elif self.args["global_"] in [GLOBAL_STANDARD, GLOBAL_WITH_PRE]:
            msg = "Global fitting mode with automatic peak detection is not supported yet."
            raise KeyError(msg)

    def define_parameters_auto(self) -> None:  # noqa: C901
        """Auto define the model parameters for local fitting."""
        positions, properties = self.__autodetect__()
        if (
            not isinstance(self.args["autopeak"], bool)
            and "modeltype" in self.args["autopeak"]
        ):
            _model = self.args["autopeak"]["modeltype"].lower()
            ReferenceKeys().automodel_check(model=_model)
            models = _model
        else:
            models = "gaussian"

        if models == "gaussian":
            for i, (_cent, _amp, _fhmw) in enumerate(
                zip(
                    self.x[positions],
                    properties["peak_heights"],
                    properties["widths"],
                ),
                start=1,
            ):
                self.params.add(
                    f"{models}_amplitude_{i}",
                    value=_amp,
                    min=-np.abs(1.25 * _amp),
                    max=np.abs(1.25 * _amp),
                    vary=True,
                )
                self.params.add(
                    f"{models}_center_{i}",
                    value=_cent,
                    min=0.5 * _cent,
                    max=2 * _cent,
                    vary=True,
                )
                self.params.add(
                    f"{models}_fwhmg_{i}",
                    value=_fhmw,
                    min=0,
                    max=2 * _fhmw,
                    vary=True,
                )
        elif models == "orcagaussian":
            for i, (_cent, _amp, _width) in enumerate(
                zip(
                    self.x[positions],
                    properties["peak_heights"],
                    properties["widths"],
                ),
                start=1,
            ):
                self.params.add(
                    f"{models}_amplitude_{i}",
                    value=_amp,
                    min=-np.abs(1.25 * _amp),
                    max=np.abs(1.25 * _amp),
                    vary=True,
                )
                self.params.add(
                    f"{models}_center_{i}",
                    value=_cent,
                    min=0.5 * _cent,
                    max=2 * _cent,
                    vary=True,
                )
                self.params.add(
                    f"{models}_width_{i}",
                    value=_width,
                    min=0,
                    max=2 * _width,
                    vary=True,
                )
        elif models == "lorentzian":
            for i, (_cent, _amp, _fhmw) in enumerate(
                zip(
                    self.x[positions],
                    properties["peak_heights"],
                    properties["widths"],
                ),
                start=1,
            ):
                self.params.add(
                    f"{models}_amplitude_{i}",
                    value=_amp,
                    min=-np.abs(1.25 * _amp),
                    max=np.abs(1.25 * _amp),
                    vary=True,
                )
                self.params.add(
                    f"{models}_center_{i}",
                    value=_cent,
                    min=0.5 * _cent,
                    max=2 * _cent,
                    vary=True,
                )
                self.params.add(
                    f"{models}_fwhml_{i}",
                    value=_fhmw,
                    min=0,
                    max=2 * _fhmw,
                    vary=True,
                )
        elif models == "voigt":
            for i, (_cent, _amp, _fhmw) in enumerate(
                zip(
                    self.x[positions],
                    properties["peak_heights"],
                    properties["widths"],
                ),
                start=1,
            ):
                self.params.add(
                    f"{models}_amplitude_{i}",
                    value=_amp,
                    min=-np.abs(1.25 * _amp),
                    max=np.abs(1.25 * _amp),
                    vary=True,
                )
                self.params.add(
                    f"{models}_center_{i}",
                    value=_cent,
                    min=0.5 * _cent,
                    max=2 * _cent,
                    vary=True,
                )
                self.params.add(
                    f"{models}_fwhmv_{i}",
                    value=_fhmw,
                    min=0,
                    max=2 * _fhmw,
                    vary=True,
                )
        elif models == "pseudovoigt":
            for i, (_cent, _amp, _fhmw) in enumerate(
                zip(
                    self.x[positions],
                    properties["peak_heights"],
                    properties["widths"],
                ),
                start=1,
            ):
                self.params.add(
                    f"{models}_amplitude_{i}",
                    value=_amp,
                    min=-np.abs(1.25 * _amp),
                    max=np.abs(1.25 * _amp),
                    vary=True,
                )
                self.params.add(
                    f"{models}_center_{i}",
                    value=_cent,
                    min=0.5 * _cent,
                    max=2 * _cent,
                    vary=True,
                )
                self.params.add(
                    f"{models}_fwhmg_{i}",
                    value=0.5 * _fhmw,
                    min=0,
                    max=_fhmw,
                    vary=True,
                )
                self.params.add(
                    f"{models}_fwhml_{i}",
                    value=0.5 * _fhmw,
                    min=0,
                    max=2 * _fhmw,
                    vary=True,
                )

        self.args["auto_generated_models"] = {
            "models": {
                key: {
                    "value": self.params[key].value,
                    "min": self.params[key].min,
                    "max": self.params[key].max,
                    "vary": self.params[key].vary,
                }
                for key in self.params
            },
            "positions": positions.tolist(),
            "properties": {key: value.tolist() for key, value in properties.items()},
        }

    def define_parameters(self) -> None:
        """Define the input parameters for a `params`-dictionary for classic fitting."""
        for key_1, value_1 in self.args["peaks"].items():
            self.define_parameters_loop(key_1=key_1, value_1=value_1)

    def define_parameters_loop(self, key_1: str, value_1: dict[str, Any]) -> None:
        """Loop through the input parameters for a `params`-dictionary.

        Args:
            key_1 (str): The key of the first level of the input dictionary.
            value_1 (Dict[str, Any]): The value of the first level of the input
                 dictionary.

        """
        for key_2, value_2 in value_1.items():
            self.define_parameters_loop_2(key_1=key_1, key_2=key_2, value_2=value_2)

    def define_parameters_loop_2(
        self,
        key_1: str,
        key_2: str,
        value_2: dict[str, Any],
    ) -> None:
        """Loop through the input parameters for a `params`-dictionary.

        Args:
            key_1 (str): The key of the first level of the input dictionary.
            key_2 (str): The key of the second level of the input dictionary.
            value_2 (Dict[str, Any]): The value of the first level of the input
                 dictionary.

        """
        for key_3, value_3 in value_2.items():
            self.define_parameters_loop_3(
                key_1=key_1,
                key_2=key_2,
                key_3=key_3,
                value_3=value_3,
            )

    def define_parameters_loop_3(
        self,
        key_1: str,
        key_2: str,
        key_3: str,
        value_3: dict[str, Any],
    ) -> None:
        """Loop through the input parameters for a `params`-dictionary.

        Args:
            key_1 (str): The key of the first level of the input dictionary.
            key_2 (str): The key of the second level of the input dictionary.
            key_3 (str): The key of the third level of the input dictionary.
            value_3 (Dict[str, Any]): The value of the third level of the input
                 dictionary.

        """
        self.params.add(f"{key_2}_{key_3}_{key_1}", **value_3)

    def define_parameters_global(self) -> None:
        """Define the input parameters for a `params`-dictionary for global fitting."""
        for col_i in range(self.col_len):
            for key_1, value_1 in self.args["peaks"].items():
                for key_2, value_2 in value_1.items():
                    for key_3, value_3 in value_2.items():
                        self._define_parameter(
                            col_i=col_i,
                            key_1=key_1,
                            key_2=key_2,
                            key_3=key_3,
                            value_3=value_3,
                        )

    def _define_parameter(
        self,
        col_i: int,
        key_1: str,
        key_2: str,
        key_3: str,
        value_3: dict[str, Any],
    ) -> None:
        """Define the input parameters for a `params`-dictionary for global fitting.

        Args:
            col_i (int): The column index.
            key_1 (str): The key of the first level of the input dictionary.
            key_2 (str): The key of the second level of the input dictionary.
            key_3 (str): The key of the third level of the input dictionary.
            value_3 (Dict[str, Any]): The value of the third level of the input
                 dictionary.

        """
        if col_i:
            if key_3 != "amplitude":
                self.params.add(
                    f"{key_2}_{key_3}_{key_1}_{col_i + 1}",
                    expr=f"{key_2}_{key_3}_{key_1}_1",
                )
            else:
                self.params.add(
                    f"{key_2}_{key_3}_{key_1}_{col_i + 1}",
                    **value_3,
                )

        else:
            self.params.add(f"{key_2}_{key_3}_{key_1}_1", **value_3)

    def define_parameters_global_pre(self) -> None:
        """Define the input parameters for a `params`-dictionary for global fitting.

        !!! warning "About `params` for global fitting"

            `define_parameters_global_pre` requires fully defined `params`-dictionary
            in the json, toml, or yaml file input. This means:

            1. Number of the spectra must be defined.
            2. Number of the peaks must be defined.
            3. Number of the parameters must be defined.
            4. The parameters must be defined.
        """
        for key_1, value_1 in self.args["peaks"].items():
            for key_2, value_2 in value_1.items():
                for key_3, value_3 in value_2.items():
                    for key_4, value_4 in value_3.items():
                        self.params.add(f"{key_3}_{key_4}_{key_2}_{key_1}", **value_4)


class SolverModels(ModelParameters):
    """Solving models for 2D and 3D data sets.

    !!! hint "Solver Modes"
        * `"2D"`: Solve 2D models via the classic `lmfit` function.
        * `"3D"`: Solve 3D models via global git. For the `global-fitting` procedure,
             the `lmfit` function is used to solve the models with an extended set of
             parameters.
          the `lmfit` function is used.
    """

    def __init__(self, df: pd.DataFrame, args: dict[str, Any]) -> None:
        """Initialize the solver modes.

        Args:
            df (pd.DataFrame): DataFrame containing the input data (`x` and `data`).
            args (Dict[str, Any]): The input file arguments as a dictionary with
                 additional information beyond the command line arguments.

        """
        super().__init__(df=df, args=args)
        self.args_solver = SolverModelsAPI(**args).model_dump()
        self.args_global = GlobalFittingAPI(**args).model_dump()
        self.params = self.return_params

    def __call__(self) -> tuple[Minimizer, Any]:
        """Solve the fitting model.

        Returns:
            Tuple[Minimizer, Any]: Minimizer class and the fitting results.

        """
        if self.args_global["global_"]:
            minimizer = Minimizer(
                self.solve_global_fitting,
                params=self.params,
                fcn_args=(self.x, self.data),
                **self.args_solver["minimizer"],
            )
        else:
            minimizer = Minimizer(
                self.solve_local_fitting,
                params=self.params,
                fcn_args=(self.x, self.data),
                **self.args_solver["minimizer"],
            )

        result = minimizer.minimize(
            **self.args_solver["optimizer"],
        )
        self.args_solver["optimizer"]["max_nfev"] = minimizer.max_nfev
        return minimizer, result

    @staticmethod
    def solve_local_fitting(
        params: dict[str, Parameters],
        x: NDArray[np.float64],
        data: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Solving the fitting problem.

        Args:
            params (Dict[str, Parameters]): The best optimized parameters of the fit.
            x (NDArray[np.float64]): `x`-values of the data.
            data (NDArray[np.float64]): `y`-values of the data as 1d-array.

        Returns:
            NDArray[np.float64]: The best-fitted data based on the proposed model.

        """
        val = np.zeros(x.shape)
        peak_kwargs: dict[tuple[str, str], Parameters] = defaultdict(dict)
        for model_name, param_value in params.items():
            _model = model_name.lower()
            ReferenceKeys().model_check(model=_model)
            c_name = _model.split("_")

            model_key = c_name[0]
            param_name = c_name[1]
            peak_id = c_name[2]
            peak_kwargs[(model_key, peak_id)][param_name] = param_value

        for key, _kwarg in peak_kwargs.items():
            val += getattr(DistributionModels(), key[0])(x, **_kwarg)
        return np.array(val - data, dtype=np.float64)

    @staticmethod
    def solve_global_fitting(
        params: dict[str, Parameters],
        x: NDArray[np.float64],
        data: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        r"""Solving the fitting for global problem.

        !!! note "About implemented models"
            `solve_global_fitting` is the global solution of `solve_local_fitting` a
            wrapper function for the calling the implemented moldels. For the kind of
            supported models see `solve_local_fitting`.

        !!! note "About the global solution"
            The global solution is a solution for the problem, where the `x`-values is
            the energy, but the y-values are the intensities, which has to be fitted as
            one unit. For this reason, the residual is calculated as the difference
            between all the y-values and the global proposed solution. Later the
            residual has to be flattened to a 1-dimensional array and minimized by the
            `lmfit`-optimizer.


        Args:
            params (Dict[str, Parameters]): The best optimized parameters of the fit.
            x (NDArray[np.float64]): `x`-values of the data.
            data (NDArray[np.float64]): `y`-values of the data as 2D-array.

        Returns:
            NDArray[np.float64]: The best-fitted data based on the proposed model.

        """
        val = np.zeros(data.shape)
        peak_kwargs: dict[tuple[str, str, str], Parameters] = defaultdict(dict)

        for model, value in params.items():
            model_lower = model.lower()
            ReferenceKeys().model_check(model=model_lower)
            c_name = model_lower.split("_")
            peak_kwargs[(c_name[0], c_name[2], c_name[3])][c_name[1]] = value
        for key, _kwarg in peak_kwargs.items():
            i = int(key[2]) - 1
            val[:, i] += getattr(DistributionModels(), key[0])(x, **_kwarg)

        val -= data
        return val.flatten()


def calculated_model(
    params: dict[str, Parameters],
    x: NDArray[np.float64],
    df: pd.DataFrame,
    global_fit: int,
) -> pd.DataFrame:
    r"""Calculate the single contributions of the models and add them to the dataframe.

    !!! note "About calculated models"
        `calculated_model` are also wrapper functions similar to `solve_model`. The
        overall goal is to extract from the best parameters the single contributions in
        the model. Currently, `lmfit` provides only a single model, so the best-fit.

    Args:
        params (Dict[str, Parameters]): The best optimized parameters of the fit.
        x (NDArray[np.float64]): `x`-values of the data.
        df (pd.DataFrame): DataFrame containing the input data (`x` and `data`),
             as well as the best fit and the corresponding residuum. Hence, it will be
             extended by the single contribution of the model.
        global_fit (int): If 1 or 2, the model is calculated for the global fit.

    Returns:
        pd.DataFrame: Extended dataframe containing the single contributions of the
            models.

    """
    peak_kwargs: dict[Any, Parameters] = defaultdict(dict)

    for model, value in params.items():
        model_lower = model.lower()
        ReferenceKeys().model_check(model=model_lower)
        p_name = model_lower.split("_")
        if global_fit:
            peak_kwargs[(p_name[0], p_name[2], p_name[3])][p_name[1]] = value
        else:
            peak_kwargs[(p_name[0], p_name[2])][p_name[1]] = value

    _df = df.copy()
    for key, _kwarg in peak_kwargs.items():
        c_name = f"{key[0]}_{key[1]}_{key[2]}" if global_fit else f"{key[0]}_{key[1]}"
        _df[c_name] = getattr(DistributionModels(), key[0])(x, **_kwarg)

    return _df


@dataclass(frozen=True)
class Constants:
    """Constants used for calculations.

    This class provides mathematical constants used across the package.
    It's implemented as a frozen dataclass with class variables
    to ensure they can't be modified.
    """

    ln2: ClassVar[float] = log(2.0)
    sq2pi: ClassVar[float] = sqrt(2.0 * pi)
    sqpi: ClassVar[float] = sqrt(pi)
    sq2: ClassVar[float] = sqrt(2.0)
    fwhmg2sig: ClassVar[float] = 1 / (2.0 * sqrt(2.0 * log(2.0)))
    fwhml2sig: ClassVar[float] = 1 / 2.0
    fwhmv2sig: ClassVar[float] = 1 / 3.60131
